__all__ = ('TransportLayerBase', 'SocketTransportLayer', 'DatagramSocketTransportLayer',)

import reprlib, selectors
import socket as module_socket
from collections import deque

from ...utils import copy_docs, include

from .abstract import AbstractTransportLayerBase
from .extra_info import (
    EXTRA_INFO_NAME_PEER_NAME, EXTRA_INFO_NAME_SOCKET, EXTRA_INFO_NAME_SOCKET_NAME, get_extra_info, has_extra_info,
    set_extra_info
)


write_exception_async = include('write_exception_async')


if hasattr(module_socket, 'TCP_NODELAY'):
    def _set_nodelay(socket):
        if (
            (socket.family in (module_socket.AF_INET, module_socket.AF_INET6)) and
            (socket.type == module_socket.SOCK_STREAM) and
            (socket.proto == module_socket.IPPROTO_TCP)
        ):
            socket.setsockopt(module_socket.IPPROTO_TCP, module_socket.TCP_NODELAY, 1)
else:
    def _set_nodelay(socket):
        pass


MAX_SIZE = 262144

class TransportLayerBase(AbstractTransportLayerBase):
    """
    Defines basic transport layer functionality.
    
    Attributes
    ----------
    _extra : `None`, `dict` of (`str`, `Any`) items
        Optional transport information.
    _loop : ``EventThread``
        The event loop to what the transport is bound to.
    """
    __slots__ = ('_extra', '_loop',)
    
    def __new__(cls, loop, extra):
        """
        Creates a new transport layer instance.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop to what the transport is bound to.
        extra : `None`, `dict` of (`str`, `Any`) items
            Optional transport information.
        """
        self = object.__new__(cls)
        self._extra = extra
        self._loop = loop
        return self
    
    @copy_docs(AbstractTransportLayerBase.get_extra_info)
    def get_extra_info(self, name, default=None):
        return get_extra_info(self._extra, name, default)
    
    
    def _fatal_error(self, exception, message='Fatal error on transport layer'):
        """
        If a fatal error occurs on the transport, renders its traceback and closes itself.
        
        Parameters
        ----------
        exception : `BaseException`
            The occurred exception.
        message : `str` = `'Fatal error on transport layer'`, Optional
            Additional error message to render.
        """
        if not isinstance(exception, (BrokenPipeError, ConnectionResetError, ConnectionAbortedError)):
            write_exception_async(
                exception,
                [
                    message,
                    '\nException occurred at \n',
                    repr(self),
                    '.\n',
                ],
                loop = self._loop,
            )


class SocketTransportLayerBase(TransportLayerBase):
    """
    Defines commonly used logic between ``SocketTransportLayer`` and ``DatagramSocketTransportLayer``.
    
    Attributes
    ----------
    _extra : `dict` of (`str`, `Any`) items
        Optional transport information.
    _loop : ``EventThread``
        The event loop to what the transport is bound to.
    _closing : `bool`
        Whether the transport ic closing.
    _connection_lost : `bool`
        Set as `True`, when ``._call_connection_lost`` is scheduled.
    _high_water : `int`
        The ``.protocol`` is paused writing when the buffer size passes the high water mark. Defaults to `65536`.
    _low_water : `int`
        The ``.protocol`` is resumed writing when the buffer size goes under the low water mark. Defaults to `16384`.
    _paused : `bool`
        Whether the transport's reading is paused by the protocol.
    _protocol : `None`, ``SSLProtocol, ``ReadProtocolBase``, `Any`
        Asynchronous protocol implementation used by the transport.
        
        After closing the transport is set to `None`.
    _protocol_paused : `bool`
        Whether ``.protocol`` is paused writing.
    _socket : `socket.socket`
        The socket used by the transport.
    _socket_file_descriptor : `int`
        The transport's socket's file descriptor identifier.
    """
    __slots__ = (
        '_closing', '_connection_lost', '_high_water', '_low_water', '_paused', '_protocol', '_protocol_paused',
        '_socket', '_socket_file_descriptor'
    )
    
    def __new__(cls, loop, extra, socket, protocol, waiter):
        """
        Creates a socket based transport layer.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop to what the transport is bound to.
        extra : `None`, `dict` of (`str`, `Any`) items
            Optional transport information.
        socket : `socket.socket`
            The socket used by the transport.
        protocol : ``ProtocolBase``
            Asynchronous protocol implementation used by the transport.
        waiter : `None`, ``Future`
            Waiter, what's result is set, when the transport connected. Defaults to `None`.
        """
        extra = set_extra_info(extra, EXTRA_INFO_NAME_SOCKET, socket)
        
        try:
            socket_name = socket.getsockname()
        except OSError:
            socket_name = None
        
        extra = set_extra_info(extra, EXTRA_INFO_NAME_SOCKET_NAME, socket_name)
        
        if has_extra_info(extra, EXTRA_INFO_NAME_PEER_NAME):
            try:
                peer_name = socket.getpeername()
            except module_socket.error:
                peer_name = None
            
            extra = set_extra_info(extra, EXTRA_INFO_NAME_PEER_NAME, peer_name)
        
        
        self = TransportLayerBase.__new__(cls, loop, extra)
        
        self._protocol_paused = False
        self._set_write_buffer_limits()
        
        self._socket = socket
        self._socket_file_descriptor = socket.fileno()
        self._protocol = protocol
        self._buffer = bytearray()
        self._connection_lost = False  # Set when call to connection_lost scheduled.
        self._closing = False  # Set when close() called.
        self._paused = False
        
        return self
    
    
    def __repr__(self):
        """Returns the socket transport layer's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        if self._socket is None:
            repr_parts.append(' closed')
        elif self._closing:
            repr_parts.append(' closing')
        
        repr_parts.append(' file_descriptor=')
        repr_parts.append(repr(self._socket_file_descriptor))
        
        loop = self._loop
        #is the transport open?
        if (loop is not None) and loop.running:
        
            try:
                key = loop.selector.get_key(self._socket_file_descriptor)
            except KeyError:
                polling = 0
            else:
                polling = key.events & selectors.EVENT_READ
            
            repr_parts.append(' read=')
            if polling:
                state = 'polling'
            else:
                state = 'idle'
            repr_parts.append(state)

            try:
                key = loop.selector.get_key(self._socket_file_descriptor)
            except KeyError:
                polling = 0
            else:
                polling = key.events & selectors.EVENT_WRITE

            repr_parts.append(' write=<')
            if polling:
                state = 'polling'
            else:
                state = 'idle'
            repr_parts.append(state)

            repr_parts.append(', buffer_size=')
            
            buffer_size = self.get_write_buffer_size()
            repr_parts.append(str(buffer_size))
            repr_parts.append('>')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __del__(self):
        """
        Closes the socket transport layer's ``.socket`` if not yet closed.
        """
        socket = self._socket
        if (socket is not None):
            self._socket = None
            socket.close()
    
    
    @copy_docs(TransportLayerBase.get_protocol)
    def get_protocol(self):
        return self._protocol
    
    
    @copy_docs(TransportLayerBase.set_protocol)
    def set_protocol(self, protocol):
        self._protocol = protocol
    
    
    @copy_docs(TransportLayerBase.is_closing)
    def is_closing(self):
        return self._closing
    
    
    @copy_docs(TransportLayerBase.close)
    def close(self):
        if self._closing:
            return
        
        self._closing = True
        self._loop.remove_reader(self._socket_file_descriptor)
        if not self._buffer:
            self._connection_lost = True
            self._loop.remove_writer(self._socket_file_descriptor)
            self._loop.call_soon(self._call_connection_lost, None)
    
    
    @copy_docs(TransportLayerBase.abort)
    def abort(self):
        self._force_close(None)
    
    
    @copy_docs(TransportLayerBase.set_write_buffer_limits)
    def set_write_buffer_limits(self, high=None, low=None):
        self._set_write_buffer_limits(high=high, low=low)
        self._maybe_pause_protocol()
    
    
    def _call_connection_lost(self, exception):
        """
        Calls the transport's connection lost method-
        
        Parameters
        ----------
        exception : `None`, ``BaseException``
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        try:
            protocol = self._protocol
            if (protocol is not None):
                protocol.connection_lost(exception)
        finally:
            socket = self._socket
            if (socket is not None):
                self._socket = None
                socket.close()
            
            self._protocol = None
    
    
    @copy_docs(TransportLayerBase.get_write_buffer_limits)
    def get_write_buffer_limits(self):
        return self._low_water, self._high_water
    
    
    def _maybe_pause_protocol(self):
        """
        Called after data was ensured to be written into the socket to check whether it's protocol should be paused.
        """
        size = self.get_write_buffer_size()
        if size <= self._high_water:
            return
        
        if self._protocol_paused:
            return
            
        self._protocol_paused = True
        
        protocol = self._protocol
        if (protocol is not None):
            try:
                protocol.pause_writing()
            except BaseException as err:
                write_exception_async(
                    err,
                    [
                        'Exception occurred at:\n',
                        repr(self),
                        '._maybe_pause_protocol\n',
                    ],
                    loop = self._loop,
                )
    
    
    def _set_write_buffer_limits(self, high=None, low=None):
        """
        Sets the write buffer limits of the transport.
        
        Parameters
        ----------
        low : None`, `int` = `None`, Optional
            The ``.protocol`` is paused writing when the buffer size passes the high water mark. Defaults to `65536`.
        high : `None`, `int` = `None`, Optional
            The ``.protocol`` is resumed writing when the buffer size goes under the low water mark. Defaults to
            `16384`.

        Raises
        ------
        ValueError
            If `high` is lower than `low` or if `low` is lower than `0`.
        """
        if high is None:
            if low is None:
                high = 65536
                low = 16384
            else:
                high = low << 2
        else:
            if low is None:
                low = high >> 2
        
        if low < 0 or high < low:
            raise ValueError(
                f'High water must be greater or equal than low, what must be greater than equal than `0`, '
                f'got high={high!r}; low={low!r}.'
            )
        
        self._high_water = high
        self._low_water = low
    

    def _force_close(self, exception):
        """
        Closes the transport immediately.
        
        The buffered data will be lost.
        
        Parameters
        ----------
        exception : `None`, ``BaseException``
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        if self._connection_lost:
            return
        
        buffer = self._buffer
        if buffer:
            buffer.clear()
            self._loop.remove_writer(self._socket_file_descriptor)
        
        if not self._closing:
            self._closing = True
            self._loop.remove_reader(self._socket_file_descriptor)
        
        self._connection_lost = True
        self._loop.call_soon(self._call_connection_lost, exception)
    
    
    def _read_ready(self):
        """
        Added as a read callback on the respective event loop to be called when the data is received on the pipe.
        
        If this happens, since it is a write only pipe, means it should be closed, so we do like that.
        """
        pass
    
    
    def _maybe_resume_protocol(self):
        """
        Called after successful writing to the socket to check whether the protocol should be resumed.
        """
        if (not self._protocol_paused) or (self.get_write_buffer_size() > self._low_water):
            return
        
        self._protocol_paused = False
        protocol = self._protocol
        if (protocol is not None):
            try:
                protocol.resume_writing()
            except BaseException as err:
                write_exception_async(
                    err,
                    [
                        'Exception occurred at:\n',
                        repr(self),
                        '._maybe_resume_protocol\n',
                    ],
                    loop = self._loop,
                )


class SocketTransportLayer(SocketTransportLayerBase):
    """
    Socket transport layer.
    
    Attributes
    ----------
    _extra : `dict` of (`str`, `Any`) items
        Optional transport information.
    _loop : ``EventThread``
        The event loop to what the transport is bound to.
    _closing : `bool`
        Whether the transport ic closing.
    _connection_lost : `bool`
        Set as `True`, when ``._call_connection_lost`` is scheduled.
    _high_water : `int`
        The ``.protocol`` is paused writing when the buffer size passes the high water mark. Defaults to `65536`.
    _low_water : `int`
        The ``.protocol`` is resumed writing when the buffer size goes under the low water mark. Defaults to `16384`.
    _paused : `bool`
        Whether the transport's reading is paused by the protocol.
    _protocol : `None`, ``SSLProtocol, ``ReadProtocolBase``, `Any`
        Asynchronous protocol implementation used by the transport.
        
        After closing the transport is set to `None`.
    _protocol_paused : `bool`
        Whether ``.protocol`` is paused writing.
    _socket : `socket.socket`
        The socket used by the transport.
    _socket_file_descriptor : `int`
        The transport's socket's file descriptor identifier.
    _at_eof : `bool`
        Whether ``.write_eof`` was called.
    _buffer : `bytearray`
        Transport's buffer.
    _server : `None`, ``Server``
        If the transport is server side, it's server is set as this attribute.
    """
    __slots__ = ('_at_eof', '_buffer', '_server')
    
    def __new__(cls, loop, extra, socket, protocol, waiter, server):
        """
        Creates a new socket transport layer.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop to what the transport is bound to.
        extra : `None`, `dict` of (`str`, `Any`) items
            Optional transport information.
        socket : `socket.socket`
            The socket used by the transport.
        protocol : ``ProtocolBase``
            Asynchronous protocol implementation used by the transport.
        waiter : `None`, ``Future`
            Waiter, what's result is set, when the transport connected. Defaults to `None`.
        server : `None`, ``Server``
            If the transport is server side, it's server is set as this attribute. Defaults to `None`.
        """
        self = SocketTransportLayerBase.__new__(cls, loop, extra, socket, protocol, waiter)
        
        self._buffer = bytearray()
        self._server = server
        self._at_eof = False
        
        if (server is not None):
            server._attach()
        
        loop.call_soon(protocol.connection_made, self)
        
        _set_nodelay(socket)
        
        # only start reading when connection_made() has been called
        loop.call_soon(loop.add_reader, self._socket_file_descriptor, self._read_ready)
        if (waiter is not None):
            # only wake up the waiter when connection_made() has been called
            loop.call_soon(type(waiter).set_result_if_pending, waiter, None)
        
        return self
    
    
    @copy_docs(SocketTransportLayerBase.write)
    def write(self, data):
        if not isinstance(data, (bytes, bytearray, memoryview)):
            raise TypeError(
                f'`data` can be `bytes-like`, got {data.__class__.__name__}; {reprlib.repr(data)}.'
            )
        
        if self._at_eof:
            raise RuntimeError(
                f'Cannot call `.write` after `.write_eof`; self={self!r}.'
            )
        
        if not data:
            return
        
        if self._connection_lost:
            return
        
        if not self._buffer:
            # Optimization: try to send now.
            try:
                n = self._socket.send(data)
            except (BlockingIOError, InterruptedError):
                pass
            
            except BaseException as err:
                self._fatal_error(err, 'Fatal write error on socket transport')
                return
            
            else:
                data = data[n:]
                if not data:
                    return
            
            # Not all was written; register write handler.
            self._loop.add_writer(self._socket_file_descriptor, self._write_ready)
        
        # Add it to the buffer.
        self._buffer.extend(data)
        self._maybe_pause_protocol()
    
    # `.writelines` same as `SocketTransportLayerBase`'s.
    
    @copy_docs(SocketTransportLayerBase.write_eof)
    def write_eof(self):
        if self._at_eof:
            return
        
        self._at_eof = True
        
        if not self._buffer:
            self._socket.shutdown(module_socket.SHUT_WR)
    

    @copy_docs(SocketTransportLayerBase.can_write_eof)
    def can_write_eof(self):
        return True
    
    
    @copy_docs(SocketTransportLayerBase.get_write_buffer_size)
    def get_write_buffer_size(self):
        return len(self._buffer)
    
    
    @copy_docs(SocketTransportLayerBase.pause_reading)
    def pause_reading(self):
        if self._closing or self._paused:
            return False
        
        self._paused = True
        self._loop.remove_reader(self._socket_file_descriptor)
        return True
    
    
    @copy_docs(SocketTransportLayerBase.resume_reading)
    def resume_reading(self):
        if not self._paused:
            return False
        
        self._paused = False
        if not self._closing:
            self._loop.add_reader(self._socket_file_descriptor, self._read_ready)
        
        return True
    
    
    
    @copy_docs(SocketTransportLayerBase._fatal_error)
    def _fatal_error(self, exception, message='Fatal error on transport layer'):
        SocketTransportLayerBase._fatal_error(self, exception, message)
        
        self._force_close(exception)
    
    
    @copy_docs(SocketTransportLayerBase._call_connection_lost)
    def _call_connection_lost(self, exception):
        try:
            SocketTransportLayerBase._call_connection_lost(self, exception)
        finally:
            server = self._server
            if (server is not None):
                self._server = None
                server._detach()
    
    
    @copy_docs(SocketTransportLayerBase._read_ready)
    def _read_ready(self):
        if self._connection_lost:
            return
        
        try:
            data = self._socket.recv(MAX_SIZE)
        except (BlockingIOError, InterruptedError):
            pass
        
        except BaseException as err:
            self._fatal_error(err, 'Fatal read error on socket transport')
            
        else:
            if data:
                self._protocol.data_received(data)
            
            elif self._protocol.eof_received():
                # We're keeping the connection open so the protocol can write more, but we still can't receive more,
                # so remove the reader callback.
                self._loop.remove_reader(self._socket_file_descriptor)
            
            else:
                self.close()
    
    
    def _write_ready(self):
        """
        Added as a write callback on the respective event loop when the transport has unsent data. Called when the
        respective socket becomes writable.
        """
        if self._connection_lost:
            return
        
        try:
            bytes_sent = self._socket.send(self._buffer)
        except (BlockingIOError, InterruptedError):
            pass
        except BaseException as err:
            self._loop.remove_writer(self._socket_file_descriptor)
            self._buffer.clear()
            self._fatal_error(err, 'Fatal write error on socket transport')
        else:
            if bytes_sent:
                del self._buffer[:bytes_sent]
            
            self._maybe_resume_protocol()  # May append to buffer.
            if not self._buffer:
                self._loop.remove_writer(self._socket_file_descriptor)
                
                if self._closing:
                    self._call_connection_lost(None)
                
                elif self._at_eof:
                    self._socket.shutdown(module_socket.SHUT_WR)


class DatagramSocketTransportLayer(SocketTransportLayerBase):
    """
    Asynchronous transport implementation for datagram sockets.
    
    Attributes
    ----------
    _extra : `dict` of (`str`, `Any`) items
        Optional transport information.
    _loop : ``EventThread``
        The event loop to what the transport is bound to.
    _closing : `bool`
        Whether the transport ic closing.
    _connection_lost : `bool`
        Set as `True`, when ``._call_connection_lost`` is scheduled.
    _high_water : `int`
        The ``.protocol`` is paused writing when the buffer size passes the high water mark. Defaults to `65536`.
    _low_water : `int`
        The ``.protocol`` is resumed writing when the buffer size goes under the low water mark. Defaults to `16384`.
    _paused : `bool`
        Whether the transport's reading is paused by the protocol.
    _protocol : `None`, ``SSLProtocol, ``ReadProtocolBase``, `Any`
        Asynchronous protocol implementation used by the transport.
        
        After closing the transport is set to `None`.
    _protocol_paused : `bool`
        Whether ``.protocol`` is paused writing.
    _socket : `socket.socket`
        The socket used by the transport.
    _socket_file_descriptor : `int`
        The transport's socket's file descriptor identifier.
    _address : `None` or (`str`, `int`)
        The last address, where the transport sent data. Defaults to `None`. The send target address should not differ
        from the last, where the transport sent data.
    """
    __slots__ = ('_address', '_buffer', )
    
    def __new__(cls, loop, extra, socket, protocol, waiter, address):
        """
        Creates a datagram transport layer with the given parameters.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop to what the transport is bound to.
        extra : `None`, `dict` of (`str`, `Any`) items
            Optional transport information.
        socket : `socket.socket`
            The socket used by the transport.
        protocol : ``ProtocolBase``
            Asynchronous protocol implementation used by the transport.
        waiter : `None`, ``Future`
            Waiter, what's result is set, when the transport connected. Defaults to `None`.
        address : `None`, `tuple` (`str`, `int`)
            The last address, where the transport sent data. The send target address should not
            differ from the last, where the transport sent data.
        """
        self = SocketTransportLayerBase.__new__(cls, loop, extra, socket, protocol, waiter)
        
        self._buffer = deque()
        self._address = address
        
        loop.call_soon(self._protocol.connection_made, self)
        # only start reading when connection_made() has been called
        loop.call_soon(self._add_reader)
        
        if (waiter is not None):
            # only wake up the waiter when connection_made() has been called
            loop.call_soon(type(waiter).set_result_if_pending, waiter, None)
        
        return self
    
    
    # `write` inherited from `TransportLayerBase`.
    
    @copy_docs(SocketTransportLayerBase.writelines)
    def writelines(self, lines):
        pass
    
    # `write_eof` inherited from `TransportLayerBase`.
    
    
    # `can_write_eof` inherited from `TransportLayerBase`.
    
    @copy_docs(SocketTransportLayerBase.get_write_buffer_size)
    def get_write_buffer_size(self):
        size = 0
        for data, address in self._buffer:
            size += len(data)
        
        return size
    
    
    # `pause_reading` inherited from `TransportLayerBase`
    
    # `resume_reading` inherited from `TransportLayerBase`
    
    
    @copy_docs(SocketTransportLayerBase._fatal_error)
    def _fatal_error(self, exception, message='Fatal error on transport'):
        if not isinstance(exception, OSError):
            write_exception_async(
                exception,
                [
                    message,
                    '\nException occurred at ',
                    repr(self),
                    '.\n',
                ],
                loop = self._loop
            )
        
        self._force_close(exception)
    
    
    def _add_reader(self):
        """
        Call soon callback added by ``.__init__` to add reader to the event loop.
        """
        if self._closing:
            return
        
        self._loop.add_reader(self._socket_file_descriptor, self._read_ready)
    
    
    @copy_docs(SocketTransportLayerBase._read_ready)
    def _read_ready(self):
        if self._connection_lost:
            return
        
        try:
            data, address = self._socket.recvfrom(MAX_SIZE)
        except (BlockingIOError, InterruptedError):
            pass
        
        except OSError as err:
            self._protocol.error_received(err)
        
        except BaseException as err:
            self._fatal_error(err, 'Fatal read error on datagram transport')
        
        else:
            self._protocol.datagram_received(data, address)
    
    
    def send_to(self, data, maybe_address=None):
        """
        Sends the given data to the target address. If target is already set, can be the parameter can be ignored.
        
        Parameters
        ----------
        data : `bytes-like`
            The data to send.
        maybe_address : `None`, `tuple` (`str`, `int`) = `None`, Optional
            The address to send the data to.
        
        Raises
        ------
        TypeError
            If `data` was not given as `bytes-like`.
        ValueError
            If `address` was given but it is different from the currently set one.
        """
        if not isinstance(data, (bytes, bytearray, memoryview)):
            raise TypeError(
                f'`data` can be `bytes-like`, got {data.__class__.__name__}; {reprlib.repr(data)}.'
            )
        
        if not data:
            return
        
        address = self._address
        if (address is not None):
            if (maybe_address is not None) or (maybe_address != address):
                raise ValueError(
                    f'Invalid address: `address` should be `None`, got {maybe_address!r}; current={address!r}.'
                )
            
            maybe_address = address
        
        if self._connection_lost:
            return
        
        buffer = self._buffer
        if not buffer:
            # Attempt to send it right away first.
            try:
                if get_extra_info(self._extra, EXTRA_INFO_NAME_PEER_NAME, None) is None:
                    self._socket.sendto(data, maybe_address)
                else:
                    self._socket.send(data)
            except (BlockingIOError, InterruptedError):
                self._loop.add_writer(self._socket_file_descriptor, self._send_to_ready)
            
            except OSError as err:
                self._protocol.error_received(err)
                return
            
            except BaseException as err:
                self._fatal_error(err, 'Fatal write error on datagram transport')
                return
            
            else:
                return
        
        # Ensure that what we buffer is immutable.
        buffer.append((bytes(data), maybe_address))
        self._maybe_pause_protocol()
    
    
    def _send_to_ready(self):
        """
        Added callback by `.send_to` to the respective event loop, when the socket is not ready for writing.
        
        This method tries to send the data again.
        """
        buffer = self._buffer
        while buffer:
            data, address = buffer.popleft()
            try:
                if get_extra_info(self._extra, EXTRA_INFO_NAME_PEER_NAME, None) is None:
                    self._socket.sendto(data, address)
                else:
                    self._socket.send(data)
            except (BlockingIOError, InterruptedError):
                buffer.appendleft((data, address))  # Try again later.
                break
            
            except OSError as err:
                self._protocol.error_received(err)
                return
            
            except BaseException as err:
                self._fatal_error(err, 'Fatal write error on datagram transport')
                return
        
        self._maybe_resume_protocol() # May append to buffer.
        if not buffer:
            self._loop.remove_writer(self._socket_file_descriptor)
            if self._closing:
                self._call_connection_lost(None)



__all__ = ('WaitContinuously', )

from ...utils import ignore_frame

from ..exceptions import CancelledError, InvalidStateError

from .future import FUTURE_STATE_CANCELLED, FUTURE_STATE_FINISHED, FUTURE_STATE_PENDING, FUTURE_STATE_RETRIEVED
from .wait_till_first import WaitTillFirst


ignore_frame(__spec__.origin, 'get_result', 'raise exception',)

class WaitContinuously(WaitTillFirst):
    """
    A future subclass, which allows waiting for future or task results continuously, yielding `1` result, when awaited.
    
    Attributes
    ----------
    _blocking : `bool`
        Whether the future is already being awaited, so it blocks the respective coroutine.
    _callbacks : `list` of `callable`
        The callbacks of the future, which are queued up on the respective event loop to be called, when the future is
        finished. These callback should accept `1` parameter, the future itself.
        
        Note, if the future is already done, then the newly added callbacks are queued up instantly on the respective
        event loop to be called.
    
    _exception : `None`, `BaseException`
        The exception set to the future as it's result. Defaults to `None`.
    _loop : ``EventThread``
        The loop to what the created future is bound.
    _result : `tuple` (`set` of ``Future``, `set` of ``Future``)
        The result of the future. Defaults to `None`.
    _state : `str`
        The state of the future.
        
        Can be set as one of the following:
        
        +---------------------------+-----------+
        | Respective name           | Value     |
        +===========================+===========+
        | FUTURE_STATE_PENDING      | `0`       |
        +---------------------------+-----------+
        | FUTURE_STATE_CANCELLED    | `1`       |
        +---------------------------+-----------+
        | FUTURE_STATE_FINISHED     | `2`       |
        +---------------------------+-----------+
        | FUTURE_STATE_RETRIEVED    | `3`       |
        +---------------------------+-----------+
        
        Note, that states are checked by memory address and not by equality. Also ``FUTURE_STATE_RETRIEVED`` is used only if
        `__debug__` is set as `True`.
    
    _callback : ``._wait_callback``
        Callback added to the waited futures.
    _last_done : `None`, ``Future``
        The last done future or task of the ``WaitContinuously``.
    """
    __slots__ = ('_last_done',)
    
    def __new__(cls, futures, loop):
        """
        Creates a new ``WaitContinuously`` from the given `futures` bound to the given `loop`.
        
        Parameters
        ----------
        futures : `None`, `iterable` of ``Future``
            The futures from which the ``WaitContinuously`` will yield the done ones. Can be given as `None`.
        loop : ``EventThread``
            The loop to what the created future will be bound to.
        """
        if (futures is None):
            pending = set()
        else:
            pending = set(futures)
        
        done = set()
        
        self = object.__new__(cls)
        self._loop = loop
        
        callback = cls._wait_callback(self)
        self._callback = callback
        
        self._result = (done, pending)
        self._exception = None
        
        self._callbacks = []
        self._blocking = False
        
        self._last_done = None
        
        if pending:
            try:
                for future in pending:
                    future.add_done_callback(callback)
            finally:
                self._state = FUTURE_STATE_PENDING
        else:
            self._state = FUTURE_STATE_FINISHED
        
        return self
    
    # `__repr__` is same as ``Future.__repr__``
    
    class _wait_callback:
        """
        ``WaitContinuously``'s callback put on the future's waited by it.
        
        Attributes
        ----------
        _parent : ``WaitContinuously``
            The parent future.
        """
        __slots__ = ('_parent',)
        
        def __init__(self, parent):
            """
            Creates a new ``WaitContinuously`` callback object with the given parent ``WaitContinuously``.
            
            Parameters
            ----------
            parent : ``WaitContinuously``
                The parent future.
            """
            self._parent = parent
        
        def __call__(self, future):
            """
            The callback, which runs when a waited future is done.
            
            Removes the done future from the parent's `pending` futures and puts it on the `done` ones. Also marks the
            parent ``WaitContinuously`` as finished if it is pending.
            
            Parameters
            ----------
            future : ``Future``
                A done waited future.
            """
            parent = self._parent
            if parent is None:
                return
            
            done, pending = parent._result
            
            pending.remove(future)
            done.add(future)
            
            if parent._state == FUTURE_STATE_PENDING:
                parent._state = FUTURE_STATE_FINISHED
                parent._last_done = future
                parent._loop._schedule_callbacks(parent)
    
    def add(self, future):
        """
        Adds a new future to the ``WaitContinuously`` to wait for.
        
        Parameters
        ----------
        future : ``Future``
            A task or future to the ``WaitContinuously`` to wait for.
        
        Raises
        ------
        RuntimeError
            If the ``WaitContinuously`` has exception set, or if it is cancelled already. At this case the
            added `future` is cancelled instantly.
        """
        state = self._state
        if state == FUTURE_STATE_PENDING:
            if future.is_done():
                self._result[0].add(future)
                self._state = FUTURE_STATE_FINISHED
                self._loop._schedule_callbacks(self)
                self._last_done = future
                return
            
            pending = self._result[1]
            pending_count = len(pending)
            pending.add(future)
            if pending_count != pending:
                future.add_done_callback(self._callback)
            
            return
        
        if state == FUTURE_STATE_FINISHED:
            exception = self._exception
            if (exception is not None):
                future.cancel()
                raise RuntimeError(
                    f'`{self.__class__.__name__}.add` called, when `{self.__class__.__name__}` is already '
                    f'finished with an exception; exception={exception!r}.'
                ) from exception
            
            if future.is_done():
                self._result[0].add(future)
                return
            
            pending = self._result[1]
            pending_count = len(pending)
            pending.add(future)
            if pending_count != len(pending):
                future.add_done_callback(self._callback)
            
            self._state = FUTURE_STATE_PENDING
            return
        
        if __debug__:
            if state == FUTURE_STATE_RETRIEVED:
                exception = self._exception
                if (exception is not None):
                    future.cancel()
                    raise RuntimeError(
                        f'`{self.__class__.__name__}.add` called, when `{self.__class__.__name__}` is already '
                        f'finished with an exception; exception={exception!r}.'
                    ) from exception
                
                if future.is_done():
                    self._result[0].append(future)
                    self._state = FUTURE_STATE_FINISHED
                    return
                
                pending = self._result[1]
                pending_count = len(pending)
                pending.add(future)
                if pending_count != len(pending):
                    future.add_done_callback(self._callback)
                
                self._state = FUTURE_STATE_PENDING
                return
        
        future.cancel()
        
        raise RuntimeError(
            f'`{self.__class__.__name__}.add` called, when `{self.__class__.__name__}` is already '
           f'cancelled; self={self!r}.'
        )
    
    def _mark_as_finished(self):
        """
        Marks self as finished, ensures it's callbacks, and cancels all the added futures.
        """
        self._state = FUTURE_STATE_FINISHED
        self._loop._schedule_callbacks(self)
        
        self._callback._parent = None
        
        done, pending = self._result
        # silence them
        if __debug__:
            for future in done:
                future.cancel()
        
        callback = self._callback
        callback._parent = None
        
        while pending:
            future = pending.pop()
            future.remove_done_callback(callback)
            future.cancel()
            done.add(future)
    
    # `futures_done` same as ``WaitTillFirst.futures_done``
    # `futures_pending` same as ``WaitTillFirst.futures_pending``
    
    def cancel(self):
        """
        Cancels the ``WaitContinuously``.
        
        By calling it, also cancels all the waited futures as well.
        
        Returns
        -------
        cancelled : `int` (`0`, `1`)
            If the future is already done, returns `0`, if it got cancelled, returns `1`-
        
        Notes
        -----
        If `__debug__` is set as `True`, then `.cancel()` also marks the future and all the waited-done futures as well
        as retrieved, causing them to not render non-retrieved exceptions.
        """
        state = self._state
        if state == FUTURE_STATE_CANCELLED:
            return 0
        
        if (self._exception is not None):
            if __debug__:
                if state == FUTURE_STATE_FINISHED:
                    self._state = FUTURE_STATE_RETRIEVED
                    return 0
                
                if state == FUTURE_STATE_RETRIEVED:
                    return 0
            
            else:
                if state == FUTURE_STATE_FINISHED:
                    return 0
        
        done, pending = self._result
        # silence them
        if __debug__:
            for future in done:
                future.cancel()
        
        callback = self._callback
        callback._parent = None
        
        while pending:
            future = pending.pop()
            future.remove_done_callback(callback)
            future.cancel()
            done.add(future)
        
        if state == FUTURE_STATE_PENDING:
            self._loop._schedule_callbacks(self)
            self._state = FUTURE_STATE_CANCELLED
        else:
            if __debug__:
                if state == FUTURE_STATE_FINISHED:
                    self._state = FUTURE_STATE_RETRIEVED
            
        return 1
    
    # `is_cancelled` is same as ``Future.is_cancelled``
    # `is_done` is same as ``Future.is_done``
    # `is_pending` is same as ``Future.is_pending``
    
    def get_result(self):
        """
        Returns the last done future.
        
        Returns
        -------
        last_done : `None`, ``Future``.
            The last future, what finished. Defaults to `None`.
        
        Raises
        ------
        CancelledError
            The future is cancelled.
        InvalidStateError
            The futures is not done yet.
        TypeError
            The future has non `BaseException` set as exception.
        BaseException
            The future's set exception.
        """
        state = self._state
        
        if state == FUTURE_STATE_FINISHED:
            if __debug__:
                self._state = FUTURE_STATE_RETRIEVED
            
            exception = self._exception
            if (exception is not None):
                raise exception
            
            return self._last_done
        
        if __debug__:
            if state == FUTURE_STATE_RETRIEVED:
                exception = self._exception
                if (exception is not None):
                    raise exception
                
                return self._last_done
        
        if state == FUTURE_STATE_CANCELLED:
            raise CancelledError
        
        # still pending
        raise InvalidStateError(self, 'result')
    
    def reset(self):
        """
        Resets the future if applicable enabling it to yield the next done future.
        
        Returns
        -------
        reset : `int` (`0`, `1`, `2`, `3`)
            - Returns `0` if the future is unable to reset. For example, there is an exception set to it, or it is
                cancelled.
            - If the future has nothing more to await on, returns `1`.
            - If there are futures to wait for, and there is no other done yet, returns `2`.
            - If there are futures to wait for, and there is at least one already other done, returns `3`.
        """
        state = self._state
        if state == FUTURE_STATE_FINISHED:
            if (self._exception is not None):
                return 0
            
            done, pending = self._result
            if done:
                last_done = self._last_done
                if (last_done is not None):
                    try:
                        done.remove(last_done)
                    except KeyError:
                        pass
            
            if done:
                self._last_done = done.pop()
                return 3
            
            if pending:
                self._state = FUTURE_STATE_PENDING
                return 2
            
            return 1
        
        if __debug__:
            if state == FUTURE_STATE_RETRIEVED:
                if (self._exception is not None):
                    return 0
                
                done, pending = self._result
                last_done = self._last_done
                if (last_done is not None):
                    try:
                        done.remove(last_done)
                    except KeyError:
                        pass
                
                if done:
                    self._state = FUTURE_STATE_FINISHED
                    self._last_done = done.pop()
                    return 3
                
                if pending:
                    self._state = FUTURE_STATE_PENDING
                    return 2
                
                return 1
        
        if state == FUTURE_STATE_PENDING:
            if self._result[1]:
                return 2
            
            return 1
        
        # cancelled
        return 0
    
    # `exception` is same as ``Future.exception``
    # `add_done_callback` is same as ``Future.add_done_callback``
    # `remove_done_callback` is same as ``Future.remove_done_callback``
    # `set_result` is same as ``WaitTillFirst.set_result``
    # `set_result_if_pending` is same as ``WaitTillFirst.set_result_if_pending``
    
    def set_exception(self, exception):
        """
        Marks the future as done and set's it's exception. If the exception is `TimeoutError`, then will not cancel it,
        but it will yield it's next result as `None` instead.
        
        Parameters
        ----------
        exception : `BaseException`, `type<BaseException>`
            The exception to set as the future's exception.
        
        Raises
        ------
        InvalidStateError
            If the future is already done by being cancelled, or it has exception set.
        TypeError
            If `StopIteration` is given as `exception`.
        """
        state = self._state
        if (state == FUTURE_STATE_CANCELLED) or ((state >= FUTURE_STATE_FINISHED) and (self._exception is not None)):
            raise InvalidStateError(self, 'set_exception')
        
        if isinstance(exception, type):
            exception = exception()
        
        if isinstance(exception, StopIteration):
            raise TypeError(
                f'{exception} cannot be raised to a(n) `{self.__class__.__name__}`; {self!r}.'
            )
        
        if isinstance(exception, TimeoutError):
            self._state = FUTURE_STATE_FINISHED
            self._last_done = None
            self._loop._schedule_callbacks(self)
        else:
            self._mark_as_finished()
            self._exception = exception
    
    def set_exception_if_pending(self, exception):
        """
        Marks the future as done and set's it's exception. If the exception is `TimeoutError`, then will not cancel it,
        but it will yield it's next result as `None` instead. Not like ``.set_exception``, this method will not raise
        ``InvalidStateError`` if the future is already done.
        
        Parameters
        ----------
        exception : `BaseException`, `type<BaseException>`
            The exception to set as the future's exception.
        
        Returns
        ------
        set_result : `int` (`0`, `1`, `2`)
            If the future is already done, returns `0`. If `exception` is given as `TimeoutError`, returns `2`, else
            `1`.
        
        Raises
        ------
        TypeError
            If `StopIteration` is given as `exception`.
        """
        state = self._state
        if (state == FUTURE_STATE_CANCELLED) or ((state >= FUTURE_STATE_FINISHED) and (self._exception is not None)):
            return 0
        
        if isinstance(exception, type):
            exception = exception()
        
        if isinstance(exception, StopIteration):
            raise TypeError(
                f'{exception} cannot be raised to a(n) `{self.__class__.__name__}`; {self!r}.'
            )
        
        if isinstance(exception, TimeoutError):
            self._state = FUTURE_STATE_FINISHED
            self._last_done = None
            self._loop._schedule_callbacks(self)
            return 2
        
        self._mark_as_finished()
        self._exception = exception
        return 1
    
    # `__iter__` is same as ``Future.__iter__``
    # `__await__` is same as ``Future.__await__``
    # if __debug__:
    #    `__del__` is same as ``Future.__del__``
    #    `__silence__` is same as ``Future.__silence__``
    #    `__silence_callback__` is same as ``Future.__silence_callback__``
    # `cancel_handles` is same as ``Future.cancel_handles``
    # `clear` same as ``WaitTilLFirst.clear``
    # `sync_wrap` is same as ``Future.sync_wrap``
    # `async_wrap` is same as ``Future.async_wrap``



__all__ = ('WeakMap',)

from .docs import has_docs
from .removed_descriptor import RemovedDescriptor
from .weak_core import WeakReferer, add_to_pending_removals


@has_docs
class _WeakMapCallback:
    """
    Callback used by ``WeakMap``-s.
    
    Attributes
    ----------
    _parent : ``WeakReferer`` to ``WeakMap``
        The parent weak map.
    """
    __slots__ = ('_parent', )
    
    @has_docs
    def __new__(cls, parent):
        """
        Creates a new ``_WeakMapCallback`` bound to the given ``WeakMap``.
        
        Parameters
        ----------
        parent : ``WeakMap``
            The parent weak map.
        """
        parent = WeakReferer(parent)
        self = object.__new__(cls)
        self._parent = parent
        return self
    
    @has_docs
    def __call__(self, reference):
        """
        Called when an element of the respective weak map is garbage collected.
        
        Parameters
        ----------
        reference : ``WeakReferer``
            Weakreference to the respective object.
        """
        parent = self._parent()
        if parent is None:
            return
        
        if parent._iterating:
            add_to_pending_removals(parent, reference)
        else:
            try:
                dict.__delitem__(parent, reference)
            except KeyError:
                pass


@has_docs
class WeakMap(dict):
    """
    Weak map is a mix of weak dictionaries and weak sets. Can be used to retrieve an already existing weakreferenced
    value from itself.
    
    Attributes
    ----------
    _pending_removals : `None`, `set` of ``WeakReferer``
        Pending removals of the weak map if applicable.
    _iterating : `int`
        Whether the weak map is iterating and how much times.
    _callback : ``_WeakMapCallback``
        Callback added to the ``WeakMap``'s weak keys.
    
    Class Attributes
    ----------------
    MAX_REPR_ELEMENT_LIMIT : `int` = `50`
        The maximal amount of items to render by ``.__repr__``.
    
    Notes
    -----
    ``WeakMap``-s are weakreferable.
    """
    __slots__ = ('__weakref__', '_pending_removals', '_iterating', '_callback')
    
    MAX_REPR_ELEMENT_LIMIT = 50
    
    @has_docs
    def _commit_removals(self):
        """
        Commits the pending removals of the weak map if applicable.
        """
        if self._iterating:
            return
        
        pending_removals = self._pending_removals
        if pending_removals is None:
            return
        
        for reference in pending_removals:
            try:
                dict.__delitem__(self, reference)
            except KeyError:
                pass
        
        self._pending_removals = None
    
    # __class__ -> same
    
    @has_docs
    def __contains__(self, key):
        """Returns whether the weak map contains the given key."""
        try:
            reference = WeakReferer(key)
        except TypeError:
            return False
        
        return dict.__contains__(self, reference)
    
    # __delattr__ -> same
    
    @has_docs
    def __delitem__(self, key):
        """Deletes the given key from the weak map"""
        try:
            reference = WeakReferer(key)
        except TypeError:
            raise KeyError(key) from None
        
        try:
            dict.__delitem__(self, reference)
        except KeyError as err:
            raise KeyError(key) from None
    
    # __dir__ -> same
    # __doc__ -> same
    
    def __eq__(self, other):
        """returns whether the two weak maps are equal."""
        if isinstance(other, type(self)):
            return dict.__eq__(self, other)
        
        if isinstance(other, set):
            pass
        
        elif hasattr(type(other), '__iter__'):
            other = set(other)
        
        else:
            return NotImplemented
        
        self_set = set(iter(self))
        
        return self_set == other
    
    # __format__ -> same
    # __ge__ -> same
    # __getattribute__ -> same
    
    @has_docs
    def __getitem__(self, key):
        """Gets the already existing key from the weak map, which matches the given one."""
        try:
            reference = WeakReferer(key)
        except TypeError:
            raise KeyError(key) from None
        
        reference = dict.__getitem__(self, reference)
        key = reference()
        if (key is None):
            if self._iterating:
                add_to_pending_removals(self, reference)
            else:
                dict.__delitem__(self, reference)
            
            raise KeyError(key)
        
        return key
    
    
    # __gt__ -> same
    # __hash__ -> same
    
    @has_docs
    def __init__(self, iterable=None):
        """
        Creates a new ``WeakMap`` from the given iterable.
        
        Parameters
        ----------
        iterable : `None`, `iterable` = `None`, Optional
            Iterable to update the created map with.
        """
        self._pending_removals = None
        self._iterating = 0
        self._callback = _WeakMapCallback(self)
        if (iterable is not None):
            self.update(iterable)
    
    # __init_subclass__ -> same
    
    @has_docs
    def __iter__(self):
        """
        Iterates over the weak map's elements.
        
        This method is an iterable generator,
        """
        self._iterating += 1
        
        try:
            for reference in dict.__iter__(self):
                key = reference()
                if (key is None):
                    add_to_pending_removals(self, reference)
                    continue
                
                yield key
                continue
        
        finally:
            self._iterating -= 1
            self._commit_removals()
    
    # __le__ -> same
    
    @has_docs
    def __len__(self):
        """Returns the length of the weak map."""
        length = dict.__len__(self)
        pending_removals = self._pending_removals
        if (pending_removals is not None):
            length -= len(pending_removals)
        
        return length
    
    # __lt__ -> same
    
    def __ne__(self, other):
        """returns whether the two weak maps are equal."""
        if isinstance(other, type(self)):
            return dict.__ne__(self, other)
        
        if isinstance(other, set):
            pass
        
        elif hasattr(type(other), '__iter__'):
            other = set(other)
        
        else:
            return NotImplemented
        
        self_set = set(iter(self))
        
        return self_set != other
    
    # __new__ -> same
    
    @has_docs
    def __reduce__(self):
        """Reduces the map to a picklable object."""
        return (type(self), list(self))
    
    
    @has_docs
    def __reduce_ex__(self, version):
        """Reduces the map to a picklable object."""
        return type(self).__reduce__(self)
    
    
    @has_docs
    def __repr__(self):
        """Returns the weak map's representation."""
        result = [self.__class__.__name__, '({']
        if len(self):
            limit = self.MAX_REPR_ELEMENT_LIMIT
            collected = 0
            
            for reference in dict.__iter__(self):
                key = reference()
                if (key is None):
                    add_to_pending_removals(self, reference)
                    continue
                
                result.append(repr(key))
                result.append(', ')
                
                collected +=1
                if collected != limit:
                    continue
                
                leftover = len(self) - collected
                if leftover:
                    result.append('...}, ')
                    result.append(str(leftover))
                    result.append(' truncated)')
                else:
                    result[-1] = '})'
                break
            else:
                result[-1] = '})'
            
            self._commit_removals()
        else:
            result.append('})')
        
        return ''.join(result)
    
    # __setattr__ -> same
    __setitem__ = RemovedDescriptor()
    # __sizeof__ -> same
    
    __str__ = __repr__
    
    # __subclasshook__ -> same
    
    @has_docs
    def clear(self):
        """
        Clear's the weak map.
        """
        dict.clear(self)
        self._pending_removals = None
    
    
    @has_docs
    def copy(self):
        """
        Copies the weak map.
        
        Returns
        -------
        new : ``WeakMap``
        """
        new = dict.__new__(type(self))
        new._iterating = 0
        new._pending_removals = None
        new._callback = callback = _WeakMapCallback(new)
        
        for reference in dict.__iter__(self):
            key = reference()
            if (key is None):
                add_to_pending_removals(self, reference)
                continue
            
            reference = WeakReferer(key, callback)
            dict.__setitem__(new, reference, reference)
            continue
        
        self._commit_removals()
        
        return new
    
    
    @has_docs
    def get(self, key, default=None):
        """
        Gets the key of the weak map, which matches the given one.
        
        Parameters
        ----------
        key : `Any`
            A key to match.
        default : `Any` = `None`, Optional
            Default value to return if the given `key` could not be matched.
        
        Returns
        -------
        real_key : `Any`, `default`
            The matched key. If no key was matched returns the `default` value.
        """
        try:
            reference = WeakReferer(key)
        except TypeError:
            return default
        
        real_reference = dict.get(self, reference, reference)
        if real_reference is reference:
            return default
        
        real_key = real_reference()
        if (real_key is not None):
            return real_key
        
        if self._iterating:
            add_to_pending_removals(self, real_reference)
        else:
            dict.__delitem__(self, real_reference)
        
        return default
    
    
    items = RemovedDescriptor()
    keys = RemovedDescriptor()
    
    
    @has_docs
    def pop(self, key, default=...):
        """
        Pops a key from the weak map which matches the given one.
        
        Parameters
        ----------
        key : `Any`
            A key to match.
        default : `Any`, Optional
            Default value to return if the given `key` could not be matched.
        
        Returns
        -------
        real_key : `Any`, `default`
            The matched key. If no key was matched and `default` value is given, then returns that.
        
        Raises
        ------
        KeyError
            If `key` could not be matched and `default` value is was not given either.
        """
        try:
            reference = WeakReferer(key)
        except TypeError:
            pass
        else:
            real_reference = dict.pop(self, reference, ...)
            if (real_reference is not ...):
                real_key = real_reference()
                if (real_key is not None):
                    return real_key
                
                if self._iterating:
                    add_to_pending_removals(self, real_reference)
                else:
                    dict.__delitem__(self, real_reference)
        
        if default is ...:
            raise KeyError(key)
        
        return default
    
    
    popitem = RemovedDescriptor()
    setdefault = RemovedDescriptor()
    
    
    @has_docs
    def update(self, iterable):
        """
        Updates the map with the given iterable.
        
        Parameters
        ----------
        iterable : `iterable`
            The iterable to update the map with.
        
        Raises
        ------
        TypeError
            If the given value is not `iterable`, or any of it's elements is not weakreferable.
        """
        if hasattr(type(iterable), '__iter__'):
            # Make sure, we have unique elements, so convert other to set
            for element in iterable:
                self.set(element)
        
        else:
            raise TypeError(
                f'Parameter `iterable` must be an iterable, got {iterable.__class__.__name__}; {iterable!r}.'
            )
    
    
    values = RemovedDescriptor()
    
    
    @has_docs
    def set(self, key):
        """
        Sets a key to the ``WeakMap`` and then returns it. If they given key is already present in the ``WeakMap``,
        returns that instead.
        
        Parameters
        ----------
        key : `Any`
            A key to match.
        
        Returns
        -------
        real_key : `Any`
            The matched key, or the given one.
        
        Raises
        ------
        TypeError
            If `key` not supports weakreferencing.
        """
        reference = WeakReferer(key, self._callback)
        real_reference = dict.get(self, reference, None)
        if (real_reference is not None):
            real_key = real_reference()
            if (real_key is not None):
                return real_key
        
        dict.__setitem__(self, reference, reference)
        return key



__all__ = ('ignore_frame', 'render_exception_into', 'render_frames_into', 'should_ignore_frame',)

import reprlib, sys, warnings
from collections import OrderedDict
from linecache import checkcache as check_file_cache, getline as _get_file_line
from os.path import sep as PATH_SEPARATOR
from types import (
    AsyncGeneratorType as CoroutineGeneratorType, CoroutineType, FrameType, FunctionType, GeneratorType, MethodType,
    TracebackType
)

from .docs import copy_docs
from .highlight import HIGHLIGHT_TOKEN_TYPES, iter_highlight_code_lines
from .highlight.token import Token


SlotWrapperType = object.__lt__.__class__

IGNORED_FRAME_LINES = {}


class ConsoleLineInput:
    """
    A cached value of ``ConsoleLineCache``. An console line instance represents an input value.
    
    Attributes
    ----------
    _content : `str`
        The inputted content.
    _lines : `None`, `list` of `str`
        Cached splitted content.
    file_name : `str`
        The respective input's file name.
    length : `str`
        The content's length.
    """
    __slots__ = ('_content', '_lines', 'file_name','length')
    
    def __new__(cls, file_name, content):
        """
        Creates a new console line input.
        
        Parameters
        ----------
        file_name : `str`
            The file name to reference content with.
        content : `str`
            The line's content.
        """
        self = object.__new__(cls)
        self._content = content
        self._lines = None
        self.file_name = file_name
        self.length = len(content)
        return self
    
    
    def __repr__(self):
        """Returns the console line input's representation."""
        return f'<{self.__class__.__name__} file_name={self.file_name!r}, length={self.length!r}>'
    
    
    def get_lines(self):
        """
        Gets the lines of the input.
        
        Returns
        -------
        lines : `list` of `str`
        """
        lines = self._lines
        if (lines is None):
            content = self._content
            # should not happen
            if (content is None):
                lines = []
            else:
                lines = content.splitlines()
                self._content = None
        
            self._lines = lines
        
        return lines
    
    
    def get_line(self, line_number):
        """
        Tries to get the line of the input.
        
        Parameters
        ----------
        line_number : `int`
            The line's number (1 based). 
        
        Returns
        -------
        line : `str`
        """
        if line_number < 1:
            return ''
        
        lines = self.get_lines()
        
        if line_number > len(lines):
            return ''
        
        return lines[line_number - 1] + '\n'


class ConsoleLineCache:
    """
    Line cash, which can be for console, so line lookup wont fail for their lines.
    
    Attributes
    ----------
    max_size : `int`
        The maximal size to cache.
    actual_size : `int`
        The actual stored size.
    cache : `OrderedDict` of (`str`, ``ConsoleLineInput``) items
        cache used to store inputs.
    """
    def __new__(cls, max_size = 1000000):
        """
        Creates a new console line cache instance.
        
        parameters
        ----------
        max_size : `int` = `1000_000`, Optional
            The maximal amount of input to remember.
        """
        self = object.__new__(cls)
        
        self.max_size = max_size
        self.actual_size = 0
        self.cache = OrderedDict()
        
        return self
    
    def __repr__(self):
        """Returns the console line cache's representation."""
        return f'<{self.__class__.__name__} size: {self.actual_size!r} / {self.max_size!r}>'
    
    
    def feed(self, file_name, content):
        """
        Feeds content with the given file-name.
        
        Feeding a new content with an old file name overwrites the old value, so make sure to always fee with a new
        name.
        
        Parameters
        ----------
        file_name : `str`
            The file name to feed content with.
        content : `str`
            The content to feed.
        """
        line_input = ConsoleLineInput(file_name, content)
        actual_size = self.actual_size + line_input.length
        max_size = self.max_size
        cache = self.cache
        
        if max_size > actual_size:
            to_pop = []
            
            for value in cache.values():
                actual_size -= value.length
                
                if actual_size <= max_size:
                    break
                
                to_pop.append(value)
                continue
            
            for value in to_pop:
                del cache[value.file_name]
        
        cache[file_name] = line_input
        self.actual_size = actual_size
    
    
    def get_line(self, file_name, line_number):
        """
        Gets the line of the given file for the given index.
        
        Parameters
        ----------
        file_name : `str`
            The file to get it's line of.
        line_number : `int`
            The line's number (1 based). 
        
        Returns
        -------
        line : `str`
        """
        cache = self.cache
        
        try:
            line_input = cache[file_name]
        except KeyError:
            return ''
        
        cache.move_to_end(file_name)
        return line_input.get_line(line_number)


CONSOLE_LINE_CACHE = ConsoleLineCache()


def get_file_line(file_name, line_number):
    """
    Tries to get the line of the file.
    
    Parameters
    ----------
    file_name : `str`
        The file to get it's line of.
    line_number : `int`
        The line's number (1 based). 
    
    Returns
    -------
    line : `str`
    """
    # First check file lines
    line = _get_file_line(file_name, line_number, None)
    
    # Second check console lines
    if not line:
        line = CONSOLE_LINE_CACHE.get_line(file_name, line_number)
    
    return line


def try_get_raw_exception_representation(exception):
    """
    Tries to get raw exception representation.
    
    Parameters
    ----------
    exception : ``BaseException``
        The respective exception instance.
    
    Returns
    -------
    raw_exception_representation : `str`
    """
    raw_exception_representation_parts = [
        '> repr(exception) raised, trying to get raw representation.\n'
    ]
    
    exception_name = getattr(type(exception), '__name__')
    if type(exception_name) is str:
        pass
    elif isinstance(exception_name, str):
        try:
            exception_name = str(exception_name)
        except:
            exception_name = '<Exception>'
    else:
        exception_name = '<Exception>'
    
    raw_exception_representation_parts.append(exception_name)
    raw_exception_representation_parts.append('(')
    
    try:
        args = getattr(exception, 'args', None)
    except:
        pass
    else:
        if (args is not None) and (type(args) is tuple):
            length = len(args)
            if length:
                index = 0
                while True:
                    element = args[index]
                    
                    try:
                        element_representation = repr(element)
                    except:
                        element_representation = f'<parameter_{index}>'
                    else:
                        if type(element_representation) is not str:
                            try:
                                element_representation = str(element_representation)
                            except:
                                element_representation = f'<parameter_{index}>'
                    
                    raw_exception_representation_parts.append(element_representation)
                    
                    index += 1
                    if index == length:
                        break
                    
                    raw_exception_representation_parts.append(', ')
                    continue
        
    
    raw_exception_representation_parts.append(')')
    return ''.join(raw_exception_representation_parts)


def is_syntax_error(exception):
    """
    Returns whether the given exception is a syntax error with the expected structure.
    
    Parameters
    ----------
    exception : ``BaseException``
        The exception to check.
    
    Returns
    -------
    is_syntax_error : `bool`
    """
    if not isinstance(exception, SyntaxError):
        return False
    
    exception_parameters = exception.args
    if (not isinstance(exception_parameters, tuple)) or (len(exception_parameters) != 2):
        return False
    
    message, details = exception_parameters
    if (message is not None) and (not isinstance(message, str)):
        return False
    
    if (not isinstance(details, tuple)) or (len(details) not in (4, 6)):
        return False
    
    # Pulled from C 3.11
    #
    #     args = Py_BuildValue("(O(OiiNii))", errmsg, tok->filename, tok->lineno,
    #                          col_offset, errtext, tok->lineno, end_col_offset);
    #
    # file_name is annotated as `O`, so it probably can be `None` as well.
    # also: line_number == end_line_number, so we can ignore it.
    #
    # On older Python versions `line` can also be `None`. How stupid!
    
    file_name = details[0]
    line_number = details[1]
    offset = details[2]
    line = details[3]
    
    if (file_name is not None) and (not isinstance(file_name, str)):
        return False
    
    if not isinstance(line_number, int):
        return False
    
    if not isinstance(offset, int):
        return False
    
    if (line is not None) and (not isinstance(line, str)):
        return False
    
    if len(details) == 4:
        return True
    
    end_line_number = details[4]
    end_offset = details[5]
    
    if not isinstance(end_line_number, int):
        return False
    
    if not isinstance(end_offset, int):
        return False
    
    return True


def fixup_syntax_error_line_from_buffer(syntax_error, buffer):
    """
    Tries to fix up the syntax error's missing line.
    
    Should be only called if ``is_syntax_error`` returned `True` on the given exception.
    
    Parameters
    ----------
    syntax_error : ``SyntaxError``
        Respective SyntaxError
    buffer : `list` of `str`
        Buffer containing the respective lines.
    """
    message, (file_name, line_number, offset, line, *end) = syntax_error.args
    if line is not None:
        return
    
    buffer_length = len(buffer)
    # `line_number` means `line_index + 1`
    if (line_number > buffer_length) or (line_number < 1):
        return
    
    line = buffer[line_number - 1]
    syntax_error.args = (message, (file_name, line_number, offset, line, *end))


def right_strip_syntax_error_line(syntax_error):
    """
    Right strips the syntax error's line. This can be useful when comparing two syntax errors, but one has new line
    character at the end.
    
    Should be only called if ``is_syntax_error`` returned `True` on the given exception.
    
    Parameters
    ----------
    syntax_error : `SyntaxError`
        The syntax error to strip.
    """
    message, (file_name, line_number, offset, line, *end) = syntax_error.args
    if line is None:
        return
    
    line = line.rstrip()
    
    syntax_error.args = (message, (file_name, line_number, offset, line, *end))


def _render_syntax_error_representation_into(syntax_error, into, highlighter):
    """
    Renders a syntax exception's representation.
    
    Parameters
    ----------
    syntax_error : ``SyntaxError``
        The respective exception instance.
    
    into : `list` of `str`
        The list of strings to extend.
    
    highlighter : `None`, ``HighlightFormatterContext``
        Formatter storing highlighting details.
    
    Returns
    -------
    into  : `list` of `str`
    """
    message, details = syntax_error.args
    
    file_name = details[0]
    line_number = details[1]
    offset = details[2]
    line = details[3]
    
    if len(details) == 6:
        end_offset = details[5]
    else:
        end_offset = -1
    
    if (file_name is None):
        file_name = ''
    
    into.extend(_iter_highlight_producer(_produce_file_location(file_name, line_number, ''), highlighter))
    into.append('\n')
    
    if (line is not None):
        line = line.strip()
        
        left_stripped_count = 0
        for character in line:
            if character in {' ', '\n', '\t', '\f'}:
                left_stripped_count += 1
                continue
            
            break
        
        line = line[left_stripped_count:]
        
        into.append('    ')
        if highlighter is None:
            into.append(line)
        else:
            into.extend(iter_highlight_code_lines([line], highlighter))
        
        into.append('\n')
        
        if offset - 1 >= left_stripped_count:
            into.append(' ' * (3 + offset - left_stripped_count))
            
            if end_offset == -1:
                pointer_length = 1
            else:
                pointer_length = end_offset - offset
                
                if pointer_length < 1:
                    pointer_length = 1
            
            pointer = '^'
            if pointer_length != 1:
                pointer = '^' * pointer_length
            
            
            into = _add_typed_part_into(
                HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_EXCEPTION_REPR, pointer, into, highlighter
            )
            
            into.append('\n')
    
    
    exception_representation = type(syntax_error).__name__
    
    if (message is not None) and message:
        exception_representation = f'{exception_representation}: {message}'
    
    into = _add_typed_part_into(
        HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_EXCEPTION_REPR, exception_representation, into, highlighter
    )
    
    return into


def _get_simple_exception_representation(exception):
    """
    Tries to get simple exception representation.
    
    Parameters
    ----------
    exception : `BaseException`
        The exception to get it's representation of.
    
    Returns
    -------
    exception_representation : `None`, `str`
        Returns `None` if simple representation is not available.
    """
    exception_type = type(exception)
    if FunctionType is SlotWrapperType:
        if (exception_type.__init__ is not BaseException.__init__):
            return None
    
    else:
        if exception_type.__init__.__class__ is not SlotWrapperType:
            return None
    
    exception_parameters = getattr(exception, 'args', None)
    if (exception_parameters is None) or (not isinstance(exception_parameters, tuple)) :
        return None
    
    exception_parameters_length = len(exception_parameters)
    if exception_parameters_length > 1:
        return None
    
    exception_representation = exception_type.__name__
    if (exception_parameters_length == 1):
        exception_representation = f'{exception_representation}: {exception_parameters[0]}'

    return exception_representation


def get_exception_representation(exception):
    """
    Gets the exception's representation.
    
    Parameters
    ----------
    exception : ``BaseException``
        The respective exception instance.
    
    Returns
    -------
    exception_representation : `str`
    """
    try:
        exception_representation = _get_simple_exception_representation(exception)
        if (exception_representation is None):
            exception_representation = repr(exception)
    except (KeyboardInterrupt, SystemExit):
        raise
    except:
        exception_representation = try_get_raw_exception_representation(exception)
    
    return exception_representation


def ignore_frame(file, name, line):
    """
    When rendering an exception traceback, specified frames can be added to being stopped from rendering.
    
    Parameters
    ----------
    file : `str`
        The name of the respective file.
    name : `str`
        The name of the respective function.
    line : `str`
        The respective line's stripped content.
    """
    try:
        files = IGNORED_FRAME_LINES[file]
    except KeyError:
        files = {}
        IGNORED_FRAME_LINES[file] = files

    try:
        names = files[name]
    except KeyError:
        names = set()
        files[name] = names

    names.add(line)


def _should_ignore_frame(file_name, name, line):
    """
    Returns whether the given frame should be ignored from rending.
    
    Called by ``should_ignore_frame`` before calling it's `filter`.
    
    Parameters
    ----------
    file_name : `str`
        The frame's respective file's name.
    name : `str`
        The frame's respective function's name.
    line : `str`
        The frame's respective stripped line.
    
    Returns
    -------
    should_ignore_frame : `bool`
        Whether the frame should be ignored.
    """
    try:
        files = IGNORED_FRAME_LINES[file_name]
    except KeyError:
        return False

    try:
        names = files[name]
    except KeyError:
        return False
    
    return (line in names)


def should_ignore_frame(file_name, name, line_number, line=..., filter=None):
    """
    Returns whether the given frame should be ignored from rending.
    
    Parameters
    ----------
    file_name : `str`
        The frame's respective file's name.
    name : `str`
        The frame's respective function's name.
    line_number : `int`
        The line's index where the exception occurred.
    line : `str`
        The frame's respective stripped line.
    filter : `None`, `callable` = `None`, Optional (Keyword only)
        Additional filter to check whether a frame should be shown.
        
        Called with 4 parameters:
        
        +---------------+-----------+---------------------------------------------------------------+
        | Name          | Type      | Description                                                   |
        +===============+===========+===============================================================+
        | file_name     | `str`     | The frame's file's name.                                      |
        +---------------+-----------+---------------------------------------------------------------+
        | name          | `str`     | The name of the function.                                     |
        +---------------+-----------+---------------------------------------------------------------+
        | line_number   | `int`     | The line's index of the file where the exception occurred.    |
        +---------------+-----------+---------------------------------------------------------------+
        | line          | `str`     | The line of the file.                                         |
        +---------------+-----------+---------------------------------------------------------------+
    
    Returns
    -------
    should_ignore : `bool`
    """
    if line is ...:
        line = line_number
        line_number = 0
        
        warnings.warn(
            (
                'Please call `should_ignore_frame` with `4` parameters: `file, name, line_number, line`\n'
                'The `3` parameter version is deprecated and will be removed in 2022 Jun.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
    
    if _should_ignore_frame(file_name, name, line):
        return True
    
    
    if (filter is not None) and (not filter(file_name, name, line_number, line)):
        return True
    
    return False


def format_callback(func, args=None, kwargs=None):
    """
    Formats the given callback to a more user friendly representation.
    
    Parameters
    ----------
    func : `callable`
        The callback to format.
    args : `None`, `iterable` of `Any` = `None`, Optional
        Additional parameters to call the `func` with.
    kwargs : `None`, `dict` of (`str`, `Any`) items = `None`, Optional
        Additional keyword parameters to call the `func` with.
    
    Returns
    -------
    result : `str`
        The formatted callback.
    """
    result = []
    # un-warp the wrappers
    while True:
        if not (None is args is kwargs):
            sub_result = ['(']
            if (args is not None) and args:
                for arg in args:
                    sub_result.append(reprlib.repr(arg))
                    sub_result.append(', ')
            
            if (kwargs is not None) and kwargs:
                for key, arg in kwargs.items():
                    sub_result.append(str(key)) # never trust
                    sub_result.append('=')
                    sub_result.append(reprlib.repr(arg))
                    sub_result.append(', ')
            
            if len(sub_result) > 1:
                del sub_result[-1]
            
            sub_result.append(')')
            result.append(''.join(sub_result))
        
        if hasattr(func, 'func'):
            wrapped = func.func
        
        else:
            if (type(func) is MethodType) and hasattr(type(func.__self__), 'get_coroutine'):
                coroutine = func.__self__.get_coroutine()
                coroutine_repr = getattr(coroutine, '__qualname__', None)
                if coroutine_repr is None:
                    coroutine_repr = getattr(coroutine, '__name__', None)
                    if coroutine_repr is None:
                        coroutine_repr = repr(coroutine)
                func_repr = f'<Bound method {func.__func__.__name__} of Task {coroutine_repr}>'
            else:
                func_repr = getattr(func, '__qualname__', None)
                if func_repr is None:
                    func_repr = getattr(func, '__name__', None)
                    if func_repr is None:
                        func_repr = repr(func)
            
            result.insert(0, func_repr)
            break
        
        args = getattr(func, 'args', None)
        kwargs = getattr(func, 'kwargs', None)
        func = wrapped
    
    return ''.join(result)


def _cut_file_name(file_name):
    """
    Cuts the given file name.
    
    Parameters
    ----------
    file_name : `str`
        The file name to cut.
    
    Returns
    -------
    file_name : `str`
    """
    for path in sys.path:
        if file_name.startswith(path + PATH_SEPARATOR):
            return '...' + file_name[len(path):]
    
    return file_name


def format_builtin(func):
    """
    Formats the given built-in's name.
    
    Parameters
    ----------
    func : `callable`
        The builtin to format.
    
    Returns
    -------
    result : `str`
        The formatted builtin.
    """
    # Cython or builtin
    name = getattr(func, '__qualname__', None)
    if name is None:
        name = getattr(func, '__name__', None)
        if name is None: # builtins might reach this part
            name = type(func).__name__
    
    return f'{name}()'


def format_coroutine(coroutine):
    """
    Formats the given coroutine to a more user friendly representation.
    
    Parameters
    ----------
    coroutine : `CoroutineType`, `GeneratorType`, `CoroutineGeneratorType`
        The coroutine to get representation of.
    
    Returns
    -------
    result : `str`
        The formatted coroutine.
    """
    coroutine_type = type(coroutine)
    if coroutine_type is GeneratorType:
        code = coroutine.gi_code
        frame = coroutine.gi_frame
        running = coroutine.gi_running
    
    elif coroutine_type is CoroutineType:
        code = coroutine.cr_code
        frame = coroutine.cr_frame
        running = coroutine.cr_running
    
    elif coroutine_type is CoroutineGeneratorType:
        code = coroutine.ag_code
        frame = coroutine.ag_frame
        running = coroutine.ag_running
    
    else:
        code = None
        frame = None
        running = -1
    
    if running == True:
        state = 'running'
    
    elif running == False:
        if frame is None:
            state = 'finished'
        
        else:
            state = 'suspended'
    
    else:
        state = 'unknown state'
    
    if running == -1:
        name = format_builtin(coroutine)
    else:
        name = format_callback(coroutine)
    
    if (code is None):
        location = 'unknown location'
    
    else:
        file_name = _cut_file_name(code.co_filename)
        
        if frame is None:
            line_number = code.co_firstlineno
        else:
            line_number = frame.f_lineno
        
        location = f'"{file_name}", line {line_number}'
    
    return f'<{name} from {location}; {state}>'


STRING_TYPE_NONE = 0
STRING_TYPE_SINGLE_QUOTE = 1
STRING_TYPE_DOUBLE_QUOTE = 2
STRING_TYPE_TRIPLE_SINGLE_QUOTE = 3
STRING_TYPE_TRIPLE_DOUBLE_QUOTE = 4

PARENTHESES_RELATION = {
    ')': '(',
    ']': '[',
    '}': '{',
}

def get_expression_lines(file_name, line_number):
    """
    Gets all the lines of the expression starting at the given position.
    
    Parameters
    ----------
    file_name : `file_name`
        The respective file's name.
    line_number : `int`
        The expression's first line's index.
    
    Returns
    -------
    expression_lines : `str`
        The lines of the expression.
    """
    parentheses_stack = []
    string_type = STRING_TYPE_NONE
    lines = []
    
    while True:
        line = get_file_line(file_name, line_number)
        if not line:
            break
        
        lines.append(line)
        
        parsing_failed = False
        line_escaped = False
        
        character_index = -1
        character_limit = len(line)
        
        while True:
            character_index += 1
            if character_index >= character_limit:
                break
            
            character = line[character_index]
            
            if string_type == STRING_TYPE_NONE:
                if character in {'(', '[', '{'}:
                    parentheses_stack.append(character)
                    continue
                
                if character in {')', ']', '}'}:
                    if not parentheses_stack:
                        parsing_failed = True
                        break
                    
                    
                    if parentheses_stack[-1] != PARENTHESES_RELATION[character]:
                        parsing_failed = True
                        break
                    
                    del parentheses_stack[-1]
                    continue
                
                if character == '\'':
                    if (
                        (character_index + 2 < character_limit) and
                        (line[character_index + 1] == '\'') and
                        (line[character_index + 2] == '\'')
                    ):
                        character_index += 2
                        string_type = STRING_TYPE_TRIPLE_SINGLE_QUOTE
                    else:
                        string_type = STRING_TYPE_SINGLE_QUOTE
                    
                    continue
                
                if character == '"':
                    if (
                        (character_index + 2 < character_limit) and
                        (line[character_index + 1] == '\"') and
                        (line[character_index + 2] == '\"')
                    ):
                        character_index += 2
                        string_type = STRING_TYPE_TRIPLE_DOUBLE_QUOTE
                    else:
                        string_type = STRING_TYPE_DOUBLE_QUOTE
                    
                    continue
                
                if character == '#':
                    break
                
                if character == '\\':
                    line_escaped = True
                    break
            
            else:
                if character == '\n':
                    if string_type in (STRING_TYPE_SINGLE_QUOTE, STRING_TYPE_DOUBLE_QUOTE):
                        parsing_failed = True
                        break
                    
                    continue
                
                if character == '\\':
                    character_index += 1
                    
                    if string_type in (STRING_TYPE_SINGLE_QUOTE, STRING_TYPE_DOUBLE_QUOTE):
                        if (character_index >= character_limit):
                            parsing_failed = True
                            break
                    
                    continue
                
                
                if character == '\'':
                    if string_type == STRING_TYPE_SINGLE_QUOTE:
                        string_type = STRING_TYPE_NONE
                        continue
                    
                    if string_type == STRING_TYPE_TRIPLE_SINGLE_QUOTE:
                        if (
                            (character_index + 2 < character_limit) and
                            (line[character_index + 1] == '\'') and
                            (line[character_index + 2] == '\'')
                        ):
                            character_index += 2
                            string_type = STRING_TYPE_NONE
                            continue
                    
                    continue
                
                if character == '"':
                    if string_type == STRING_TYPE_DOUBLE_QUOTE:
                        string_type = STRING_TYPE_NONE
                        continue
                    
                    if string_type == STRING_TYPE_TRIPLE_DOUBLE_QUOTE:
                        if (
                            (character_index + 2 < character_limit) and
                            (line[character_index + 1] == '"') and
                            (line[character_index + 2] == '"')
                        ):
                            character_index += 2
                            
                            string_type = STRING_TYPE_NONE
                            continue
                    
                    continue
        
        
        if parsing_failed:
            break
        
        
        if not line.endswith('\n'):
            # no more lines. Simple.
            should_continue_parsing = False
        elif string_type in (STRING_TYPE_TRIPLE_SINGLE_QUOTE, STRING_TYPE_TRIPLE_DOUBLE_QUOTE):
            should_continue_parsing = True
        elif line_escaped:
            should_continue_parsing = True
        elif parentheses_stack:
            should_continue_parsing = True
        else:
            should_continue_parsing = False
        
        
        if not should_continue_parsing:
            break
        
        line_number += 1
        continue
    
    if not lines:
        return lines
    
    for index in range(len(lines)):
        lines[index] = lines[index].rstrip()
    
    indentation_to_remove = 0
    
    while True:
        for line in lines:
            if len(line) <= indentation_to_remove:
                continue
            
            if line[indentation_to_remove] in (' ', '\t'):
                continue
            
            break
        
        else:
            indentation_to_remove += 1
            continue
        
        break
    
    
    for index in range(len(lines)):
        lines[index] = lines[index][indentation_to_remove:]
    
    return lines


class FrameDetail:
    """
    Represents information about a frame.
    
    Used when rendering frames.
    
    Attributes
    ----------
    file_name : `str`
        Path to the represented file.
    line_number : `int`
        The respective line's index.
    lines : `None`, `list` of `str`
        The python source code lines of the respective instruction.
    mention_count : `int`
        How much times this exact frame was mentioned.
    name : `str`
        The represented function's name.
    produced_content : `None`, `list` of `str`
        The produced content.
    """
    __slots__ = ('file_name', 'line_number', 'lines', 'mention_count', 'name', 'produced_content')
    
    def __new__(cls, file_name, line_number, name):
        self = object.__new__(cls)
        self.file_name = file_name
        self.name = name
        self.line_number = line_number
        self.lines = None
        self.produced_content = None
        self.mention_count = 0
        return self
    
    def __repr__(self):
        """Returns the frame info's representation."""
        return f'{self.__class__.__name__}({self.file_name!r}, {self.line_number!r}, {self.name!r})'
    
    
    def __eq__(self, other):
        """Returns whether self equals to other."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.file_name != other.file_name:
            return False
        
        if self.line_number != other.line_number:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the frame info's hash value."""
        return hash(self.file_name) ^ self.line_number ^ hash(self.name)
    
    
    def do_mention(self):
        """
        Increments how much times the frame is mentioned.
        """
        self.mention_count += 1
    
    
    def get_lines(self):
        """
        Returns the lines of the instruction of the trace.
        
        Returns
        -------
        lines : `list` of `str`
            Can be empty.
        """
        lines = self.lines
        if (lines is None):
            lines = get_expression_lines(self.file_name, self.line_number)
            self.lines = lines
        
        return lines
    
    
    def get_line(self):
        """
        Returns the line where the exception occurred.
        
        Returns
        -------
        line : `str`
        """
        return '\n'.join(self.get_lines())
    
    
    def _render(self, highlighter):
        """
        Produces each part of the frame to render.
        
        This method is an iterable generator.
        
        Parameters
        ----------
        highlighter : `None`, ``HighlightFormatterContext``
            Formatter storing highlighting details.
        
        Yields
        ------
        part : `str`
        """
        
        yield from _iter_highlight_producer(
            _produce_file_location(self.file_name, self.line_number, self.name),
            highlighter,
        )
        
        yield '\n'
        
        lines = self.get_lines()
        
        if lines:
            if (highlighter is None):
                for line in lines:
                    yield '    '
                    yield line
                    yield '\n'
            
            else:
                yield from iter_highlight_code_lines([f'    {line}\n' for line in lines], highlighter)
    
    
    def render_into(self, into, highlighter):
        """
        Renders frame.
        
        Parameters
        ----------
        into : `list` of `str`
            The list of strings to render the representation into.
        highlighter : `None`, ``HighlightFormatterContext``
            Formatter storing highlighting details.
        
        Returns
        -------
        into : `list` of `str`
        """
        if self.mention_count > 1:
            produced_content = self.produced_content
            if (produced_content is None):
                produced_content = [*self._render(highlighter)]
                self.produced_content = produced_content
            
            into.extend(produced_content)
        else:
            into.extend(self._render(highlighter))
        
        return into


FRAME_DETAIL_GROUP_TYPE_NONE = 0
FRAME_DETAIL_GROUP_TYPE_SINGLES = 1
FRAME_DETAIL_GROUP_TYPE_MAYBE_REPEAT = 2
FRAME_DETAIL_GROUP_TYPE_REPEAT = 3


class FrameDetailGroup:
    """
    Represents a group of frame details.
    
    Attributes
    ----------
    details : `list` of ``FrameDetail``
        The stored details by the group.
    group_type : `int`
        The group's type.
    repeat_count : `int`
        How much times is the group repeated.
    """
    __slots__ = ('details', 'group_type', 'repeat_count')
    
    def __new__(cls):
        """
        Creates a new frame detail group.
        """
        self = object.__new__(cls)
        self.details = []
        self.group_type = FRAME_DETAIL_GROUP_TYPE_NONE
        self.repeat_count = 1
        return self
    
    
    @classmethod
    def _create_repeated(cls, repeat_count, details):
        """
        Creates a repeated frame detail group.
        
        Parameters
        ----------
        repeat_count : `int`
            How much times is the group repeated.
        details : `list` of ``FrameDetail``
            The stored details by the group.
        
        Returns
        -------
        self : ``FrameDetailGroup``
        """
        self = object.__new__(cls)
        self.details = details
        self.group_type = FRAME_DETAIL_GROUP_TYPE_REPEAT
        self.repeat_count = repeat_count
        return self
    
    
    def __eq__(self, other):
        """Returns whether the two frame detail groups are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if len(self) != len(other):
            return False
        
        for self_detail, other_detail in zip(self.iter_details(), other.iter_details()):
            if self_detail != other_detail:
                return False
        
        return True
    
    
    def __len__(self):
        """Returns how much details are in the group."""
        return len(self.details) * self.repeat_count
    
    
    def __repr__(self):
        """Returns the frame detail group's representation."""
        return f'<{self.__class__.__name__} details={self.details!r}>'
    
    
    def __bool__(self):
        """Returns whether the detail group has any elements."""
        return self.group_type != FRAME_DETAIL_GROUP_TYPE_NONE
    
    
    def add_initial_frame_detail(self, frame_detail):
        """
        Adds the given frame detail to self. Should be only used to add the first element to self.
        
        Parameters
        ----------
        frame_detail : ``FrameDetail``
            The frame detail to add.
        """
        self.details.append(frame_detail)
        
        if frame_detail.mention_count > 1:
            group_type = FRAME_DETAIL_GROUP_TYPE_MAYBE_REPEAT
        else:
            group_type = FRAME_DETAIL_GROUP_TYPE_SINGLES
        self.group_type = group_type
    
    
    def try_add_frame_detail(self, frame_detail):
        """
        Tries to add the given frame detail to self. On success returns `True`.
        
        Parameters
        ----------
        frame_detail : ``FrameDetail``
            The frame detail to add.
        
        Returns
        -------
        success : `bool`
        """
        group_type = self.group_type
        
        if group_type == FRAME_DETAIL_GROUP_TYPE_NONE:
            self.add_initial_frame_detail(frame_detail)
            return True
        
        
        if group_type == FRAME_DETAIL_GROUP_TYPE_SINGLES:
            if frame_detail.mention_count > 1:
                return False
            
            self.details.append(frame_detail)
            return True
        
        
        if group_type == FRAME_DETAIL_GROUP_TYPE_MAYBE_REPEAT:
            self.details.append(frame_detail)
            
            if frame_detail.mention_count > 1:
                group_type = FRAME_DETAIL_GROUP_TYPE_REPEAT
            else:
                group_type = FRAME_DETAIL_GROUP_TYPE_SINGLES
            self.group_type = group_type
            return True
        
        
        if group_type == FRAME_DETAIL_GROUP_TYPE_REPEAT:
            if frame_detail.mention_count <= 1:
                return False
            
            self.details.append(frame_detail)
            return True
        
        # No more cases
        return False
    
    
    def try_merge_single_group(self, other):
        """
        Tries to merge self with an other frame detail group. But only merges them if both is possibly a single-mention
        group.
        
        Parameters
        ----------
        other : ``FrameDetailGroup``
            The detail group to merge self with.
        
        Returns
        -------
        merged : `bool`
        """
        if (self.repeat_count != 1) or (other.repeat_count != 1):
            return False
        
        self_group_type = self.group_type
        other_group_type = other.group_type
        
        if self_group_type == FRAME_DETAIL_GROUP_TYPE_SINGLES:
            should_merge = other_group_type in (FRAME_DETAIL_GROUP_TYPE_SINGLES, FRAME_DETAIL_GROUP_TYPE_MAYBE_REPEAT)
        
        elif self_group_type == FRAME_DETAIL_GROUP_TYPE_MAYBE_REPEAT:
            should_merge = other_group_type == FRAME_DETAIL_GROUP_TYPE_SINGLES
        
        else:
            should_merge = False
        
        if should_merge:
            self.details.extend(other.iter_exhaust_details())
        
        return should_merge
    
    
    def try_merge_repeat_group(self, other):
        """
        Tries to merge self with an other frame detail group. But only merges them if both is possibly a repeat-mention
        group.
        
        Parameters
        ----------
        other : ``FrameDetailGroup``
            The detail group to merge self with.
        
        Returns
        -------
        merged : `bool`
        """
        if (self.repeat_count != 1) or (other.repeat_count != 1):
            return False
        
        self_group_type = self.group_type
        other_group_type = other.group_type
        
        if self_group_type in (FRAME_DETAIL_GROUP_TYPE_MAYBE_REPEAT, FRAME_DETAIL_GROUP_TYPE_REPEAT):
            should_merge = other_group_type in (FRAME_DETAIL_GROUP_TYPE_MAYBE_REPEAT, FRAME_DETAIL_GROUP_TYPE_REPEAT)
        
        else:
            should_merge = False
        
        if should_merge:
            self.details.extend(other.iter_exhaust_details())
        
        return should_merge
    
    
    def iter_details(self):
        """
        Iterates over the details of the detail group.
        
        This method is an iterable generator.
        
        Yields
        ------
        detail : ``FrameDetail``
        """
        details = self.details
        for index in range(self.repeat_count):
            yield from details
    
    
    def iter_exhaust_details(self):
        """
        Iterates over the details of the detail group, exhausting its own.
        
        This method is an iterable generator.
        
        Yields
        ------
        detail : ``FrameDetail``
        """
        try:
            yield from self.iter_details()
        finally:
            self.details.clear()
            self.group_type = FRAME_DETAIL_GROUP_TYPE_NONE
    
    
    def iter_break_repeated_patterns(self):
        """
        Breaks down self to repeated patterns if possible.
        
        This method is an iterable generator.
        
        Yields
        ------
        detail_groups : ``FrameDetailGroup``
        """
        # Can we break it down?
        if (self.group_type != FRAME_DETAIL_GROUP_TYPE_REPEAT):
            yield self
            return
        
        # If it has repeat set, it is already a broken group
        if self.repeat_count > 1:
            yield self
            return
        
        details = self.details
        details_length = len(details)
        start = details_length - 1
        detail_groups = []
        
        
        while True:
            for reverse_shift in range(start, 1, -1):
                for chunk_length in range(1, (reverse_shift >> 1) + 2):
                    repeat = 1
                    for chunk_end_index in range(reverse_shift - chunk_length, -1, -chunk_length):
                        for index in range(chunk_length):
                            if details[reverse_shift - index] != details[chunk_end_index - index]:
                                break
                        else:
                            repeat += 1
                            continue
                        
                        break
                    
                    if repeat * chunk_length <= 2:
                        continue
                    
                    if repeat > 1:
                        if start != reverse_shift:
                            detail_groups.append(type(self)._create_repeated(1, details[reverse_shift + 1 : start + 1]))
                        
                        detail_groups.append(type(self)._create_repeated(
                            repeat, details[reverse_shift - chunk_length + 1: reverse_shift + 1]
                        ))
                        start = reverse_shift - repeat * chunk_length
                        break
                    
                else:
                    continue
                break
            else:
                break
        
        # Did we match anything even?
        if start == details_length - 1:
            yield self
            return
        
        if start != details_length:
            detail_groups.append(type(self)._create_repeated(1 , details[0 : start + 1]))
        
        yield from reversed(detail_groups)
    
    
    def render_into(self, into, highlighter):
        """
        Renders frame group.
        
        Parameters
        ----------
        into : `list` of `str`
            The list of strings to render the representation into.
        highlighter : `None`, ``HighlightFormatterContext``
            Formatter storing highlighting details.
        
        Returns
        -------
        into : `list` of `str`
        """
        repeat_count = self.repeat_count
        details = self.details
        
        if (repeat_count > 1):
            into = _add_typed_part_into(
                HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_DETAIL_FRAME_REPEAT,
                (
                    f'[Following {len(details)} frame{"s were" if len(details) > 1 else " was"} '
                    f'repeated {repeat_count} times]'
                ),
                into,
                highlighter
            )
            into.append('\n')
        
        for detail in details:
            detail.render_into(into, highlighter)
        
        if (repeat_count > 1):
            into = _add_typed_part_into(
                HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_DETAIL_FRAME_REPEAT,
                f'[End of repeated frames]',
                into,
                highlighter
            )
            into.append('\n')
        
        return into


def render_frames_into(frames, extend=None, *, filter=None, highlighter=None):
    """
    Renders the given frames into a list of strings.
    
    Parameters
    ----------
    frames : `list` of (`FrameType`, `TraceBack`, ``FrameProxyType``)
        The frames to render.
    
    extend : `None`, `list` of `str` = `None`, Optional
        Whether the frames should be rendered into an already existing list.
    
    filter : `None`, `callable` = `None`, Optional (Keyword only)
        Additional filter to check whether a frame should be shown.
        
        Called with 4 parameters:
        
        +---------------+-----------+---------------------------------------------------------------+
        | Name          | Type      | Description                                                   |
        +===============+===========+===============================================================+
        | file_name     | `str`     | The frame's file's name.                                      |
        +---------------+-----------+---------------------------------------------------------------+
        | name          | `str`     | The name of the function.                                     |
        +---------------+-----------+---------------------------------------------------------------+
        | line_number   | `int`     | The line's index of the file where the exception occurred.    |
        +---------------+-----------+---------------------------------------------------------------+
        | line          | `str`     | The line of the file.                                         |
        +---------------+-----------+---------------------------------------------------------------+
    
    highlighter : `None`, ``HighlightFormatterContext`` = `None`, Optional (Keyword only)
        Formatter storing highlighting details.
    
    Returns
    -------
    extend : `list` of `str`
        The rendered frames as a `list` of it's string parts.
    """
    frames = _convert_frames(frames)
    
    if extend is None:
        extend = []
    
    checked_file_caches = set()
    
    frame_details_ignored = set()
    frame_details_stack = []
    frame_details_unique = {}
    
    
    for frame in frames:
        frame_detail = FrameDetail(frame.file_name, frame.line_number, frame.name)
        if frame_detail in frame_details_ignored:
            continue
        
        try:
            frame_detail_to_add = frame_details_unique[frame_detail]
        except KeyError:
            
            if frame_detail.file_name not in checked_file_caches:
                checked_file_caches.add(frame_detail.file_name)
                check_file_cache(frame_detail.file_name)
                
            if should_ignore_frame(
                frame_detail.file_name,
                frame_detail.name,
                frame_detail.line_number,
                frame_detail.get_line(),
                filter = filter
            ):
                frame_details_ignored.add(frame_detail)
                continue
            
            frame_detail_to_add = frame_detail
            frame_details_unique[frame_detail_to_add] = frame_detail_to_add
        else:
            frame_detail_to_add.do_mention()
        
        frame_detail_to_add.do_mention()
        frame_details_stack.append(frame_detail_to_add)
    
    
    if len(frame_details_unique) == len(frame_details_stack):
        for frame_detail in frame_details_stack:
            extend = frame_detail.render_into(extend, highlighter)
    
    else:
        frame_detail_group = FrameDetailGroup()
        frame_detail_groups = [frame_detail_group]
        
        for frame_detail in frame_details_stack:
            if not frame_detail_group.try_add_frame_detail(frame_detail):
                frame_detail_group = FrameDetailGroup()
                frame_detail_group.add_initial_frame_detail(frame_detail)
                frame_detail_groups.append(frame_detail_group)
        
        
        for index in reversed(range(1, len(frame_detail_groups))):
            frame_detail_group_1 = frame_detail_groups[index - 1]
            frame_detail_group_2 = frame_detail_groups[index]
            
            if (
                frame_detail_group_1.try_merge_repeat_group(frame_detail_group_2) or
                frame_detail_group_1.try_merge_single_group(frame_detail_group_2)
            ):
                del frame_detail_groups[index]
                continue
        
        for frame_detail_group in frame_detail_groups:
            for frame_detail_group in frame_detail_group.iter_break_repeated_patterns():
                extend = frame_detail_group.render_into(extend, highlighter)
    
    return extend


def _produce_file_location(file_name, line_number, name):
    """
    Produces file location part and their tokens.
    
    Parameters
    ----------
    file_name : `str`
        Path of the respective file.
    line_number : int`
        The respective line's number.
    name : `str`
        The respective functions name.
    
    Yields
    ------
    part : `str`
        Part to render
    token_type : `int`
        The part's type.
    """
    yield '  File ', HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_LOCATION
    
    if file_name:
        yield f'"{file_name}"', HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_LOCATION_PATH
    else:
        yield 'unknown location', HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_LOCATION
    
    yield ', line ', HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_LOCATION
    yield  str(line_number), HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_LOCATION_LINE_NUMBER
    
    if name:
        yield ', in ', HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_LOCATION
        yield name, HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_LOCATION_NAME


class FrameProxyType:
    """
    Base class for frame proxies.
    
    Frame proxies are used to provide the same api for different types of frame representations.
    """
    __slots__ = ()
    
    def __new__(cls):
        """
        Creates a new frame proxy.
        """
        return object.__new__(cls)
    
    
    def __repr__(self):
        """Returns the frame proxy's representation."""
        return f'<{self.__class__.__name__}>'
    
    
    @property
    def builtins(self):
        """
        Returns the frame's builtins.
        
        Returns
        -------
        builtins : `dict` of (`str`, `Any`) items
        """
        raise NotImplementedError
    
    
    @property
    def globals(self):
        """
        Returns the frame's globals.
        
        Returns
        -------
        globals : `dict` of (`str`, `Any`)
        """
        raise NotImplementedError
    
    
    @property
    def locals(self):
        """
        Returns the local variables of the frame.
        
        Returns
        -------
        locals : `dict` of (`str`, `Any`) items
        """
        raise NotImplementedError
    
    
    @property
    def code(self):
        """
        Returns the frame's code.
        
        Returns
        -------
        code : `CodeType`
        """
        raise NotImplementedError
    
    
    @property
    def last_instruction_index(self):
        """
        Returns the traceback's frame's last attempted instruction index in the bytecode.
        
        Returns
        -------
        last_instruction_index : `int`
        """
        raise NotImplementedError
    
    
    @property
    def line_number(self):
        """
        Returns the traceback's frame's current line number in Python source code.
        
        Returns
        -------
        line_number : `int`
        """
        raise NotImplementedError
    
    
    @property
    def tracing_function(self):
        """
        Tracing function for the traceback's frame.
        
        Returns
        -------
        tracing_function : `FunctionType`, `None`
            Defaults to `None`.
        """
        raise NotImplementedError
    
    
    @property
    def file_name(self):
        """
        Returns the frame's file name-
        
        Returns
        -------
        file_name : `str`
        """
        file_name = self.code.co_filename
        if (file_name is None):
            file_name = ''
        
        return file_name
    
    
    @property
    def name(self):
        """
        Returns the frame's function's name.
        
        Returns
        -------
        name : `str`
        """
        return self.code.co_name


class TracebackFrameProxy(FrameProxyType):
    """
    Wraps a traceback's frame.
    
    Attributes
    ----------
    _traceback : ``TracebackType``
        The wrapped traceback frame.
    """
    __slots__ = ('_traceback',)
    
    def __new__(cls, traceback):
        """
        Creates a new traceback frame proxy.
        
        Parameters
        ----------
        traceback : ``TracebackType``
            The traceback frame to wrap.
        """
        self = object.__new__(cls)
        self._traceback = traceback
        return self
    
    
    @property
    @copy_docs(FrameProxyType.builtins)
    def builtins(self):
        return self._traceback.tb_frame.f_builtins
    
    
    @property
    @copy_docs(FrameProxyType.globals)
    def globals(self):
        return self._traceback.tb_frame.f_globals
    
    
    @property
    @copy_docs(FrameProxyType.locals)
    def locals(self):
        return self._traceback.tb_frame.f_locals
    
    
    @property
    @copy_docs(FrameProxyType.code)
    def code(self):
        return self._traceback.tb_frame.f_code
    
    
    @property
    @copy_docs(FrameProxyType.last_instruction_index)
    def last_instruction_index(self):
        return self._traceback.tb_frame.f_lasti
    
    
    @property
    @copy_docs(FrameProxyType.line_number)
    def line_number(self):
        return self._traceback.tb_lineno
    
    
    @property
    @copy_docs(FrameProxyType.tracing_function)
    def tracing_function(self):
        return self._traceback.tb_frame.f_trace



class FrameProxy(FrameProxyType):
    """
    Wraps a frame.
    
    Attributes
    ----------
    _frame : ``FrameType``
        The wrapped frame.
    """
    __slots__ = ('_frame',)
    
    def __new__(cls, frame):
        """
        Creates a new frame proxy.
        
        Parameters
        ----------
        frame : ``FrameType``
            The frame to wrap.
        """
        self = object.__new__(cls)
        self._frame = frame
        return self
    
    
    @property
    @copy_docs(FrameProxyType.builtins)
    def builtins(self):
        return self._frame.f_builtins
    
    
    @property
    @copy_docs(FrameProxyType.globals)
    def globals(self):
        return self._frame.f_globals
    
    
    @property
    @copy_docs(FrameProxyType.locals)
    def locals(self):
        return self._frame.f_locals
    
    
    @property
    @copy_docs(FrameProxyType.code)
    def code(self):
        return self._frame.f_code
    
    
    @property
    @copy_docs(FrameProxyType.last_instruction_index)
    def last_instruction_index(self):
        return self._frame.f_lasti
    
    
    @property
    @copy_docs(FrameProxyType.line_number)
    def line_number(self):
        return self._frame.f_lineno
    
    
    @property
    @copy_docs(FrameProxyType.tracing_function)
    def tracing_function(self):
        return self._frame.f_trace


def _convert_frames(frames):
    """
    Converts the given frames into frame proxies.
    
    Parameters
    ----------
    frames : `list` of `Any`
        The frames to convert.
    
    Returns
    -------
    frame_proxies : ``FrameProxy``
    
    Raises
    ------
    TypeError
        Unknown frame type.
    """
    frame_proxies = []
    
    for frame in frames:
        if isinstance(frame, FrameProxyType):
            frame_proxy = frame
        
        elif isinstance(frame, FrameType):
            frame_proxy = FrameProxy(frame)
        
        elif isinstance(frame, TracebackType):
            frame_proxy = TracebackFrameProxy(frame)
        
        else:
            raise TypeError(
                f'Unknown frame type, got {type(frame).__name__}; {frame!r}; frames={frames!r}.'
            )
        
        frame_proxies.append(frame_proxy)
    
    return frame_proxies


def _get_exception_frames(exception):
    """
    Gets the frames of the given exception.
    
    Parameters
    ----------
    exception : `BaseException`
        The exception to trace back.
    
    Returns
    -------
    frames : `list` of ``FrameProxyType``
        A list of `frame` compatible exception frames.
    """
    frames = []
    traceback = exception.__traceback__
    
    while True:
        if traceback is None:
            break
        
        frame = TracebackFrameProxy(traceback)
        frames.append(frame)
        traceback = traceback.tb_next
    
    return frames


def render_exception_into(exception, extend=None, *, filter=None, highlighter=None):
    """
    Renders the given exception's frames into a list of strings.
    
    Parameters
    ----------
    exception : `BaseException`
        The exception to render.
    
    extend : `None`, `list` of `str` = `None`, Optional
        Whether the frames should be rendered into an already existing list.
    
    filter : `None`, `callable` = `None`, Optional (Keyword only)
        Additional filter to check whether a frame should be shown.
        
        Called with 4 parameters:
        
        +---------------+-----------+---------------------------------------------------------------+
        | Name          | Type      | Description                                                   |
        +===============+===========+===============================================================+
        | file_name     | `str`     | The frame's file's name.                                      |
        +---------------+-----------+---------------------------------------------------------------+
        | name          | `str`     | The name of the function.                                     |
        +---------------+-----------+---------------------------------------------------------------+
        | line_number   | `int`     | The line's index of the file where the exception occurred.    |
        +---------------+-----------+---------------------------------------------------------------+
        | line          | `str`     | The line of the file.                                         |
        +---------------+-----------+---------------------------------------------------------------+
    
    highlighter : `None`, ``HighlightFormatterContext`` = `None`, Optional (Keyword only)
        Formatter storing highlighting details.
    
    Returns
    -------
    extend : `list` of `str`
        The rendered frames as a `list` of it's string parts.
    """
    if extend is None:
        extend = []
    
    exceptions = []
    reason_type = 0
    while True:
        exceptions.append((exception, reason_type))
        cause_exception = exception.__cause__
        if (cause_exception is not None):
            exception = cause_exception
            reason_type = 1
            continue
        
        context_exception = exception.__context__
        if (context_exception is not None):
            exception = context_exception
            reason_type = 2
            continue
        
        # no other cases
        break
    
    for exception, reason_type in reversed(exceptions):
        extend = _add_trace_title_into('Traceback (most recent call last):', extend, highlighter)
        frames = _get_exception_frames(exception)
        extend.append('\n')
        extend = render_frames_into(frames, extend, filter=filter, highlighter=highlighter)
        
        if is_syntax_error(exception):
            extend = _render_syntax_error_representation_into(exception, extend, highlighter)
        else:
            extend = _add_typed_part_into(
                HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_EXCEPTION_REPR,
                get_exception_representation(exception),
                extend,
                highlighter,
            )
        
        extend.append('\n')
        
        if reason_type == 0:
            break
        
        if reason_type == 1:
            extend.append('\n')
            extend = _add_trace_title_into(
                'The above exception was the direct cause of the following exception:', extend, highlighter
            )
            extend.append('\n\n')
            continue
        
        if reason_type == 2:
            extend.append('\n')
            extend = _add_trace_title_into(
                'During handling of the above exception, another exception occurred:', extend, highlighter
            )
            extend.append('\n\n')
            continue
        
        # no more cases
        continue
    
    return extend


def _add_trace_title_into(title, into, highlighter):
    """
    Adds trace title into the given list of strings.
    
    Parameters
    ----------
    title : `str`
        The title to add.
    into : `list` of `str`
        The list of strings to add.
    highlighter : `None`, ``HighlightFormatterContext`` = `None`, Optional (Keyword only)
        Formatter storing highlighting details.
    
    Returns
    -------
    into : `list` of `str`
    """
    return _add_typed_part_into(HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE, title, into, highlighter)


def _iter_highlight_producer(producer, highlighter):
    """
    iterates over a producer and applies highlight to each part.
    
    This method is an iterable generator
    
    Parameters
    ----------
    producer : `GeneratorType`
        Generator to produce parts.
    highlighter : `None`, ``HighlightFormatterContext``
        Formatter storing highlighting details.
    
    Yields
    ------
    part : `str`
    """
    if highlighter is None:
        for part, token_type in producer:
            yield part
    else:
        for part, token_type in producer:
            yield from highlighter.generate_highlighted(Token(token_type, part))


def _add_typed_part_into(type_, part, into, highlighter):
    """
    Adds trace title into the given list of strings.
    
    Parameters
    ----------
    type_ : `int`
        Token type identifier.
    part : `str`
        The part to add.
    into : `list` of `str`
        The list of strings to add.
    highlighter : `None`, ``HighlightFormatterContext`` = `None`, Optional (Keyword only)
        Formatter storing highlighting details.
    
    Returns
    -------
    into : `list` of `str`
    """
    if (highlighter is None):
        into.append(part)
    else:
        into.extend(highlighter.generate_highlighted(Token(type_, part)))
    
    return into



__all__ = ()

from ...utils import copy_docs

from ..protocols_and_transports import AbstractTransportLayerBase
from ..traps import skip_ready_cycle


class SubprocessWriter(AbstractTransportLayerBase):
    """
    Writer interface for subprocess calls.
    
    Attributes
    ----------
    _loop : ``EventThread``
        The respective event loop of the stream.
    _transport : ``UnixWritePipeTransport``, `Any`
        Asynchronous transport implementation.
    _protocol : ``AsyncProcess``, `Any`
        Asynchronous protocol implementation.
    """
    __slots__ = ('_loop', '_transport', '_protocol', )
    
    def __new__(cls, loop, transport, protocol):
        """
        Writer used as ``AsyncProcess.stdin``.
        
        Parameters
        ----------
        loop : ``EventThread``
            The respective event loop of the stream.
        transport : ``UnixWritePipeTransportLayer``
            Asynchronous transport implementation.
        protocol : ``AsyncProcess``
            Asynchronous protocol implementation.
        """
        self = object.__new__(cls)
        
        self._transport = transport
        self._protocol = protocol
        self._loop = loop
        
        return self
    
    
    @copy_docs(AbstractTransportLayerBase.write)
    def write(self, data):
        self._transport.write(data)
    
    
    @copy_docs(AbstractTransportLayerBase.writelines)
    def writelines(self, lines):
        self._transport.writelines(lines)
    
    
    @copy_docs(AbstractTransportLayerBase.write_eof)
    def write_eof(self):
        return self._transport.write_eof()
    
    
    @copy_docs(AbstractTransportLayerBase.can_write_eof)
    def can_write_eof(self):
        return self._transport.can_write_eof()
    
    
    @copy_docs(AbstractTransportLayerBase.close)
    def close(self):
        return self._transport.close()
    
    
    @copy_docs(AbstractTransportLayerBase.is_closing)
    def is_closing(self):
        return self._transport.is_closing()
    
    
    @copy_docs(AbstractTransportLayerBase.get_extra_info)
    def get_extra_info(self, name, default=None):
        return self._transport.get_extra_info(name, default)
    
    
    async def drain(self):
        """
        Blocks till the write buffer is drained.
        
        This method is a coroutine.
        
        Raises
        ------
        BaseException
            Connection lost exception if applicable.
        """
        if self._transport.is_closing():
            # skip 1 loop
            await skip_ready_cycle()
        
        await self._protocol._drain_helper()



__all__ = ()

from ssl import CertificateError, MemoryBIO, SSLError, SSL_ERROR_SYSCALL, SSL_ERROR_WANT_READ, SSL_ERROR_WANT_WRITE


SSL_PIPE_STATE_UNWRAPPED = 0
SSL_PIPE_STATE_DO_HANDSHAKE = 1
SSL_PIPE_STATE_WRAPPED = 2
SSL_PIPE_STATE_SHUTDOWN = 3

SSL_PIPE_STATE_VALUE_TO_NAME = {
    SSL_PIPE_STATE_UNWRAPPED: 'unwrapped',
    SSL_PIPE_STATE_DO_HANDSHAKE: 'do_handshake',
    SSL_PIPE_STATE_WRAPPED: 'wrapped',
    SSL_PIPE_STATE_SHUTDOWN: 'shutdown',
}

MAX_SIZE = 262144

class SSLPipe:
    """
    An SSL pipe.
    
    Allows you to communicate with an SSL/TLS protocol instance through memory buffers. It can be used to implement a
    security layer for an existing connection where you don't have access to the connection's file descriptor, or for
    some reason you don't want to use it.
    
    An ``SSLPipe`` can be in `wrapped` and `unwrapped` mode. In unwrapped mode, data is passed through untransformed.
    In wrapped mode, application level data is encrypted to SSL record level data and vice versa. The SSL record level
    is the lowest level in the SSL protocol suite and is what travels as-is over the wire.
    
    An ``SSLPipe`` initially is in `unwrapped` mode. To start SSL, call ``.do_handshake``, to shutdown SSL, call
    ``.unwrap``.
    
    Attributes
    ----------
    _handshake_callback : `None`, `callable`
        A Callback which will be called when handshake is completed. Set by ``.do_handshake``.
        
        Should accept the following parameters:
        +-----------------------+---------------------------+
        | Respective name       | Value                     |
        +=======================+===========================+
        | handshake_exception   | `None`, `BaseException` |
        +-----------------------+---------------------------+
        
        If the handshake is successful, then the `handshake_exception` is given as `None`, else as an exception
        instance.
    _incoming : `MemoryBIO`
        Does the incoming data encryption/decryption.
    _need_ssl_data : `bool`
        Whether more record level data is needed to complete a handshake that is currently in progress.
    _outgoing : `MemoryBIO`
        Does the outgoing data encryption/decryption.
    _server_host_name : `None`, `str`
        The ssl protocol's server hostname if applicable.
    _server_side : `bool`
        Whether the ssl protocol is server side.
    _shutdown_callback : `None`, `callable`
        A callback which will be called when the shutdown is completed. Set by ``.shutdown``.
        
        Should accept no parameters.
    _ssl_context : `SSLContext`
        The SSL pipe's SSL type.
    _ssl_object : `None`, `ssl.SSLObject`
        SSL object connecting ``._incoming`` and ``._outgoing`` memory bios set at ``.do_handshake``.
    _state : `int`
        The state of the ``SSLPipe``.
        
        Can be set as one of:
        
        +-------------------------------+-----------+
        | Respective name               | Value     |
        +===============================+===========+
        | SSL_PIPE_STATE_UNWRAPPED      | `0`       |
        +-------------------------------+-----------+
        | SSL_PIPE_STATE_DO_HANDSHAKE   | `1`       |
        +-------------------------------+-----------+
        | SSL_PIPE_STATE_WRAPPED        | `2`       |
        +-------------------------------+-----------+
        | SSL_PIPE_STATE_SHUTDOWN       | `3`       |
        +-------------------------------+-----------+
        
        Note, that ``.state`` is checked by memory address and not by value.
    """
    __slots__ = (
        '_handshake_callback', '_incoming', '_need_ssl_data', '_outgoing', '_server_host_name', '_server_side',
        '_shutdown_callback', '_ssl_context', '_ssl_object', '_state'
    )
    
    def __init__(self, context, server_side, server_host_name):
        """
        Creates a new ``SSLPipe`` with the given parameters.
        
        Parameters
        ----------
        context : `SSLContext`
            The SSL pipe's SSL type.
        server_side : `bool`
            Whether the ssl protocol is server side.
        server_host_name : `None`, `str`
            The ssl protocol's server hostname if applicable.

        """
        self._ssl_context = context
        self._server_side = server_side
        self._server_host_name = server_host_name
        self._state = SSL_PIPE_STATE_UNWRAPPED
        self._incoming = MemoryBIO()
        self._outgoing = MemoryBIO()
        self._ssl_object = None
        self._need_ssl_data = False
        self._handshake_callback = None
        self._shutdown_callback = None
    
    
    def is_wrapped(self):
        """
        Returns whether a security layer is currently is effect.
        
        In handshake or on in shutdown is always `False`.
        
        Returns
        -------
        is_wrapped : `bool`
        """
        return (self._state == SSL_PIPE_STATE_WRAPPED)
    
    
    def do_handshake(self, callback=None):
        """
        Starts the SSL handshake.
        
        Parameters
        ----------
        callback : `None`, `callable` = `None`, Optional
            A Callback which will be called when handshake is completed.
            
            Should accept the following parameters:
            +-----------------------+---------------------------+
            | Respective name       | Value                     |
            +=======================+===========================+
            | handshake_exception   | `None`, `BaseException` |
            +-----------------------+---------------------------+
            
            If the handshake is successful, then the `handshake_exception` is given as `None`, else as an exception
            instance.
            
        Returns
        -------
        ssl_data : `list` of `bytes`
            A list of SSL data. Always an empty list is returned.
        
        Raises
        ------
        RuntimeError
            If the handshake is in progress or is already completed.
        """
        if self._state != SSL_PIPE_STATE_UNWRAPPED:
            raise RuntimeError('Handshake in progress or completed.')
        
        self._ssl_object = self._ssl_context.wrap_bio(self._incoming, self._outgoing, server_side=self._server_side,
            server_hostname=self._server_host_name)
        
        self._state = SSL_PIPE_STATE_DO_HANDSHAKE
        self._handshake_callback = callback
        ssl_data, application_data = self.feed_ssl_data(b'', only_handshake=True)
        
        return ssl_data
    
    
    def shutdown(self, callback=None):
        """
        Starts the SSL shutdown process.
        
        Parameters
        ----------
        callback : `None`, `callable` = `None`, Optional
            A callback which will be called when the shutdown is completed.
            
            Should accept no parameters.
        
        Returns
        -------
        ssl_data : `list` of `bytes`
            A list of SSL data.
        
        Raises
        ------
        RuntimeError
            - If the security layer is not yet present.
            - If shutdown is already in progress.
        """
        state = self._state
        if state == SSL_PIPE_STATE_UNWRAPPED:
            raise RuntimeError('No security layer present.')
        
        if state == SSL_PIPE_STATE_SHUTDOWN:
            raise RuntimeError('Shutdown in progress.')
        
        self._state = SSL_PIPE_STATE_SHUTDOWN
        self._shutdown_callback = callback
        ssl_data, application_data = self.feed_ssl_data(b'')
        return ssl_data
    
    
    def feed_eof(self):
        """
        Sends an eof.
        
        Raises
        ------
        SSLError
            If the eof was unexpected. The raised ssl error has `errno` set as `SSL_ERROR_EOF`.
        """
        self._incoming.write_eof()
        # Ignore return.
        self.feed_ssl_data(b'')
    
    
    def feed_ssl_data(self, data, only_handshake=False):
        """
        Feed SSL record level data into the pipe.
        
        Parameters
        ----------
        data : `bytes`
            If empty `bytes` is given, then it will get ssl_data for the handshake initialization.
        
        Returns
        -------
        ssl_data : `list` of `bytes`
            Contains the SSL data that needs to be sent to the remote SSL.
        application_data : `list` of `bytes`
            Plaintext data that needs to be forwarded to the application.
            
            Might contain an empty `bytes`, what indicates that ``.shutdown`` should be called.
        """
        state = self._state
        ssl_data = []
        application_data = []
        
        if state == SSL_PIPE_STATE_UNWRAPPED:
            # If unwrapped, pass plaintext data straight through.
            if data:
                application_data.append(data)
            
            return ssl_data, application_data
        
        self._need_ssl_data = False
        if data:
            self._incoming.write(data)
        
        try:
            if state == SSL_PIPE_STATE_DO_HANDSHAKE:
                # Call do_handshake() until it doesn't raise anymore.
                self._ssl_object.do_handshake()
                self._state = state = SSL_PIPE_STATE_WRAPPED
                
                handshake_callback = self._handshake_callback
                if (handshake_callback is not None):
                    handshake_callback(None)
                
                if only_handshake:
                    return ssl_data, application_data
                # Handshake done: execute the wrapped block
            
            if state == SSL_PIPE_STATE_WRAPPED:
                # Main state: read data from SSL until close_notify.
                while True:
                    chunk = self._ssl_object.read(MAX_SIZE)
                    application_data.append(chunk)
                    if not chunk:  # close_notify
                        break
            
            elif state == SSL_PIPE_STATE_SHUTDOWN:
                # Call shutdown() until it doesn't raise anymore.
                self._ssl_object.unwrap()
                self._ssl_object = None
                self._state = SSL_PIPE_STATE_UNWRAPPED
                
                shutdown_callback = self._shutdown_callback
                if (shutdown_callback is not None):
                    shutdown_callback()
            
            elif state == SSL_PIPE_STATE_UNWRAPPED:
                # Drain possible plaintext data after close_notify.
                application_data.append(self._incoming.read())
        
        except (SSLError, CertificateError) as err:
            err_number = getattr(err, 'errno', None)
            if err_number not in (SSL_ERROR_WANT_READ, SSL_ERROR_WANT_WRITE, SSL_ERROR_SYSCALL):
                if self._state == SSL_PIPE_STATE_DO_HANDSHAKE:
                    handshake_callback = self._handshake_callback
                    if (handshake_callback is not None):
                        handshake_callback(err)
                raise
            
            self._need_ssl_data = (err_number == SSL_ERROR_WANT_READ)

        # Check for record level data that needs to be sent back. Happens for the initial handshake and renegotiation.
        if self._outgoing.pending:
            ssl_data.append(self._outgoing.read())
        
        return ssl_data, application_data
    
    
    def feed_application_data(self, data, offset):
        """
        Feed plaintext data into the pipe.
        
        If the returned `offset` is less than the returned data's total length in bytes, that case is called
        `short write`. If `short-write` happens, than the same `data` should be passed to ``.feed_application_data``
        next time. (This is an OpenSSL requirement.) A further particularity is that a short write will always have
        `offset = 0`, because the `_ssl` module does not enable partial writes. And even though the offset is `0`,
        there will still be encrypted data in ssl_data.
        
        Returns
        -------
        ssl_data : `list` of `bytes-like`
            Containing record level data that needs to be sent to the remote SSL instance.
        offset : `int`
            The number of the processed plaintext data in bytes. Might be less than the total length of the returned
            data.
        
        Raises
        ------
        SSLError
            Any unexpected SSL error.
        """
        ssl_data = []
        if self._state == SSL_PIPE_STATE_UNWRAPPED:
            # pass through data in unwrapped mode
            if offset < len(data):
                ssl_data.append(memoryview(data)[offset:])
            offset = len(data)
        else:
            view = memoryview(data)
            while True:
                self._need_ssl_data = False
                try:
                    if offset < len(view):
                        offset += self._ssl_object.write(view[offset:])
                except SSLError as err:
                    # It is not allowed to call `.write` after ``.unwrap`` until the close_notify is acknowledged. We
                    # return the condition to the caller as a short write.
                    if err.reason == 'PROTOCOL_IS_SSL_PIPE_STATE_SHUTDOWN':
                        err.errno = SSL_ERROR_WANT_READ
                        need_ssl_data = True
                    else:
                        errno = getattr(err, 'errno', None)
                        if errno is None:
                            raise
                        
                        if errno == SSL_ERROR_WANT_READ:
                            need_ssl_data = True
                        elif errno == SSL_ERROR_WANT_WRITE or errno == SSL_ERROR_SYSCALL:
                            need_ssl_data = False
                        else:
                            raise
                    
                    self._need_ssl_data = need_ssl_data
    
                # See if there's any record level data back for us.
                if self._outgoing.pending:
                    ssl_data.append(self._outgoing.read())
                
                if offset == len(view) or self._need_ssl_data:
                    break
        
        return ssl_data, offset



__all__ = ('WeakCallable', 'WeakHasher', 'WeakReferer', 'is_weakreferable', 'weak_method',)

from .compact import NEEDS_DUMMY_INIT
from .docs import has_docs
from .method_like import MethodLike

try:
    from _weakref import ref as WeakrefType
except ImportError:
    from weakref import ref as WeakrefType


@has_docs
def is_weakreferable(object_):
    """
    Returns whether the given object is weakreferable.
    
    Parameters
    ----------
    object_ : `Any`
        The object to check.
    
    Returns
    -------
    is_weakreferable : `bool`
    """
    slots = getattr(type(object_), '__slots__', None)
    if (slots is not None) and ('__weakref__' in slots):
        return True
    
    if hasattr(object_, '__dict__'):
        return True
    
    return False


@has_docs
class WeakHasher:
    """
    Object to store unhashable weakreferences.
    
    Attributes
    ----------
    _hash : `int`
        The hash of the respective reference.
    reference : ``WeakReferer``
        A dead reference to hash.
    """
    __slots__ = ('_hash', 'reference',)
    
    @has_docs
    def __init__(self, reference):
        """
        Creates a new ``WeakHasher`` from the given reference.
        
        Parameters
        ----------
        reference : ``WeakReferer``
            A dead reference to hash.
        """
        self._hash = object.__hash__(reference)
        self.reference = reference
    
    @has_docs
    def __hash__(self):
        """Returns the ``WeakHasher``'s hash value."""
        return self._hash
    
    @has_docs
    def __eq__(self, other):
        """Returns whether the two ``WeakHasher``-s are the same."""
        self_reference = self.reference
        if self_reference is other:
            return True
        
        if type(self) is not type(other):
            return NotImplemented
        
        return (self.reference is other.reference)
    
    @has_docs
    def __repr__(self):
        """Returns the ``WeakHasher``'s representation."""
        return f'{self.__class__.__name__}({self.reference!r})'
    
    @has_docs
    def __getattr__(self, name):
        """Returns the attribute of the ``WeakHasher``'s reference."""
        return getattr(self.reference, name)


@has_docs
def add_to_pending_removals(container, reference):
    """
    Adds the given weakreference to the given set.
    
    Parameters
    ----------
    container : `Any`
        The parent object, which is iterating right now, so it's items cannot be removed.
    reference : ``WeakReferer``
        The weakreference to add to the set.
    """
    try:
        hash(reference)
    except TypeError:
        reference = WeakHasher(reference)
    
    pending_removals = container._pending_removals
    if pending_removals is None:
        container._pending_removals = pending_removals = set()
    
    pending_removals.add(reference)


# speedup builtin stuff, CPython is welcome
@has_docs
class WeakReferer(WeakrefType):
    """
    Weakreferences to an object.
    
    After calling it returns the referenced object or `None` if already dead.
    """
    __slots__ = ()
    if NEEDS_DUMMY_INIT:
        def __init__(self, *args, **kwargs):
            pass
    else:
        __init__ = object.__init__
    
    
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        value = self()
        if value is None:
            repr_parts.append(' (dead)')
        else:
            repr_parts.append(' value=')
            repr_parts.append(repr(value))
        
        repr_parts.append('>')
        return ''.join(repr_parts)


del WeakrefType


@has_docs
class KeyedReferer(WeakReferer):
    """
    Weakreferences an object with a key, what can be used to identify it.
    
    Attributes
    ----------
    key : `Any`
        Key to identify the weakreferenced object.
    """
    __slots__ = ('key', )
    
    @has_docs
    def __new__(cls, obj, callback, key, ):
        """
        Creates a new ``KeyedReferer`` with the given parameters.
        
        Parameters
        ----------
        obj : `Any`
            The object to weakreference.
        callback : `Any`
            Callback running when the object is garbage collected.
        key : `Any`
            Key to identify the weakreferenced object.
        """
        self = WeakReferer.__new__(cls, obj, callback)
        self.key = key
        return self


@has_docs
class WeakCallable(WeakReferer):
    """
    Weakreferences a callable object.
    
    When the object is called, calls the weakreferenced object if not yet collected.
    """
    __slots__ = ()
    
    @has_docs
    def __call__(self, *args, **kwargs):
        """
        Calls the weakreferenced object if not yet collected.
        
        Parameters
        ----------
        *args : Parameters
            Parameters to call the weakreferenced callable with.
        **kwargs : Keyword parameters
            Keyword parameters to call the weakreferenced callable with..
        
        Returns
        -------
        result : `Any`
            The returned value by the referenced object. Returns `None` if the object is already collected.
        
        Raises
        ------
        BaseException
            Raised exception by the referenced callable.
        """
        self = WeakReferer.__call__(self)
        if self is None:
            return
        
        return self(*args, **kwargs)
    
    @has_docs
    def is_alive(self):
        """
        Returns whether the ``WeakCallable`` is still alive (the referred object by it is not collected yet.)
        
        Returns
        -------
        is_alive : `bool`
        """
        return (WeakReferer.__call__(self) is not None)


@has_docs
class weak_method(WeakReferer, MethodLike):
    """
    A method like, what weakreferences it's object not blocking it from being garbage collected.
    
    Attributes
    ----------
    __func__ : `callable`
        The function to call as a method.
    
    Class Attributes
    ----------------
    __reserved_argcount__ : `int` = `1`
        The amount of reserved parameters by weak_method.
    """
    __slots__ = ('__func__',)
    
    __reserved_argcount__ = 1
    
    @has_docs
    def __new__(cls, obj, func, callback=None):
        """
        Creates a new ``weak_method`` with the given parameter.
        
        Parameters
        ----------
        obj : `Any`
            The object to weakreference and pass to `func`.
        func : `callable`
            The function to call as a method.
        callback : `Any` = `None`, Optional
            Callback running when the object is garbage collected.
        """
        self = WeakReferer.__new__(cls, obj, callback)
        self.__func__ = func
        return self
    
    @property
    @has_docs
    def __self__(self):
        """
        Returns the weakreferenced object by the ``weak_method``. `None` if it was already garbage collected.
        
        Returns
        -------
        obj : `Any`
            The weakreferenced object if not yet garbage collected. Defaults to `None`.
        """
        return WeakReferer.__call__(self)
    
    @has_docs
    def __call__(self, *args, **kwargs):
        """
        Calls the weak_method object's function with it's object if not yet collected.
        
        Parameters
        ----------
        *args : Parameters
            Parameters to call the function with.
        **kwargs : Keyword parameters
            Keyword parameters to call the function with.
        
        Returns
        -------
        result : `Any`
            The returned value by the function. Returns `None` if the object is already collected.
        
        Raises
        ------
        BaseException
            Raised exception by the function.
        """
        obj = WeakReferer.__call__(self)
        if obj is None:
            return
        
        return self.__func__(obj, *args, **kwargs)
    
    @has_docs
    def is_alive(self):
        """
        Returns whether the ``weak_method``'s object is still alive (the referred object by it is not collected yet.)
        
        Returns
        -------
        is_alive : `bool`
        """
        return (WeakReferer.__call__(self) is not None)
    
    def __getattr__(self, name):
        return getattr(self.__func__, name)
    
    @classmethod
    @has_docs
    def from_method(cls, method_, callback=None):
        """
        Creates a new ``weak_method`` from the given `method`.
        
        Parameters
        ----------
        method_ : `method`
            The method tu turn into ``weak_method``.
        callback : `Any`, Optional
            Callback running when the respective object is garbage collected.
        """
        self = WeakReferer.__new__(cls, method_.__self__, callback)
        self.__func__ = method_.__func__
        return self



from .event_loop import *
from .flow import *
from .task import *
from .trace import *

__all__ = (
    'catching',
    
    *event_loop.__all__,
    *flow.__all__,
    *task.__all__,
    *trace.__all__,
)


# Define shortcuts

from .trace import ExceptionWriterContextManager as catching
